
import warnings
import json
import os.path as op
import numpy as np
from pandas import DataFrame
import matplotlib.pyplot as plt
from monty.json import MSONable, MontyEncoder
from pymatgen.core.periodic_table import Element
from pymatgen.analysis.phase_diagram import PhaseDiagram, GrandPotentialPhaseDiagram, PDPlotter
from pymatgen.symmetry.groups import SpaceGroup
from pynter.tools.format import format_composition
import copy


class Chempots(MSONable):
    
    def __init__(self,chempots_dict,ndecimals=6):
        """
        Class to handle set of chemical potentials. Behaves like a python dictionary.
        The dictionary needs to be set with element symbols as keys and chemical potentials 
        as values.

        Parameters
        ----------
        chempots_dict : (dict)
            Dictionary of chemical potentials in the format {el:value}.
        ndecimals : (int)
            Round the chemical potentials to this number of decimals. If None
            the numbers are left untouched. the default is 6.
            
        """
        if ndecimals:
            self._mu = {el:round(v,ndecimals) for el,v in chempots_dict.items()}
        else:
            self._mu = chempots_dict
    
    @property
    def mu(self):
        """
        Dictionary with chemical potentials
        """
        return self._mu

    def __str__(self):
        return self.mu.__str__()
    
    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.mu)

    def __iter__(self):
        return self.mu.keys().__iter__()

    def __getitem__(self,el):
        return self.mu[el]

    def __setitem__(self,el,value):
        self.mu[el] = value
        return
    
    def __delitem__(self,el):
        del self.mu[el]
        return
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.mu == other
        elif isinstance(other, Chempots):
            return self.mu == other.mu
        else:
            return False

    def keys(self):
        return self.mu.keys()

    def values(self):
        return self.mu.values()
    
    def items(self):
        return self.mu.items()

    def copy(self):
        return Chempots(copy.deepcopy(self.mu))
    
    def update(self, other):
        if isinstance(other, dict):
            for key, value in other.items():
                self.mu[key] = value
        else:
            for key, value in other:
                self.mu[key] = value
    
    
    def as_dict(self):
        """
        Json-serializable dict representation of a Chempots object. 
        
        Returns
        -------
        dict
            Json-serializable dict of a Chempots object.
        """
        d = {
        "@module": self.__class__.__module__,
        "@class": self.__class__.__name__,
        "chempots":self.mu
        }
        return d
    
    @staticmethod
    def from_dict(d):
        """
        Constructor of Chempots object from dictionary representation.
        
        Parameters
        ----------
        d : dict
        
        Returns
        -------
        Chempots object.
        """
        chempots = d["chempots"]
        return Chempots(chempots)
    
    @staticmethod
    def from_pmg_elements(d):
        """
        Construct a Chempots object from a dictionary in the format {Element:value}. Useful to
        interface with pymatgen which uses Element class to address elements instead of just their symbols.

        Parameters
        ----------
        d : (dict)
            Dictionary in the format {Element:value}.

        Returns
        -------
        Chempots.
        """
        return Chempots({el.symbol:value for el,value in d.items()})
    
    def to_pmg_elements(self):
        """
        Convert the Chempots object to a dictionary in the format {Element:value}. 
        Useful to interface with pymatgen which uses Element class to address elements 
        instead of just their symbols.

        Returns
        -------
        Dictionary in the format {Element:value}
        """
        return {Element(el):value for el,value in self.mu.items()}
    
    
    def get_absolute(self,mu_refs):
        """
        Get Chempots object with chemical potentials converted to absolute values (mu + mu_ref)

        Parameters
        ----------
        mu_refs : (dict or Chempots)
            Chempots object or just dictionary with elemental chemical potentials (mu0).

        Returns
        -------
        Chempots values converted to absolute (mu + mu_ref).
        """
        return Chempots({el:self.mu[el] + mu_refs[el] for el in self.mu})
        
    
    def get_referenced(self,mu_refs):
        """
        Get Chempots object with chemical potentials converted to referenced values (mu - mu_ref)

        Parameters
        ----------
        mu_refs : (dict or Chempots)
            Chempots object or just dictionary with elemental chemical potentials (mu0).

        Returns
        -------
        Chempots values converted to referenced (mu - mu_ref).
        """
        return Chempots({el:self.mu[el] - mu_refs[el] for el in self.mu})
            
        
        
class Reservoirs(MSONable):
    
    def __init__(self,res_dict,phase_diagram=None,mu_refs=None,are_chempots_delta=False):
        """
        Class to handle dictionaries of chemical potentials. Works almost completely like a python dictionary.

        Parameters
        ----------
        res_dict : (dict)
            Dictionary with reservoir names as key and dictionaries of chemical potentials as values.
        phase_diagram : (PhaseDiagram object), optional
            PhaseDiagram object (Pymatgen), useful to convert absolute chempots in referenced chempots. The default is None.
        mu_refs : (dict)
            Dictionary with chemical potentials of reference elements ({Element:chempot}). If the PhaseDiagram is provided
            mu_refs is taken from the mu_refs attribute. 
        are_chempots_delta : (bool), optional
            Set this variable to True if chempots in dictionary are referenced values. The default is False.
        """
        self.res_dict = res_dict
        self.pd = phase_diagram if phase_diagram else None
        if mu_refs:
            self.mu_refs = mu_refs
        elif self.pd:
            self.mu_refs = PDHandler(self.pd).get_chempots_reference()
        else:
            self.mu_refs = mu_refs
            warnings.warn('Neither PhaseDiagram or reference chempots have been provided, conversions btw ref and abs value will not be possible',UserWarning)
        self.are_chempots_delta = are_chempots_delta


    def __str__(self):
        df = self.get_dataframe()
        return df.__str__()
    
    def __repr__(self):
        return self.__str__()

    def __len__(self):
        return len(self.res_dict)

    def __iter__(self):
        return self.res_dict.keys().__iter__()

    def __getitem__(self,reskey):
        return self.res_dict[reskey]

    def __setitem__(self,reskey,chempots):
        self.res_dict[reskey] = chempots
        return
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.res_dict == other
        elif isinstance(other, Reservoirs):
            return self.res_dict == other.res_dict
        else:
            return False

    def keys(self):
        return self.res_dict.keys()

    def values(self):
        return self.res_dict.values()
    
    def items(self):
        return self.res_dict.items()

    def copy(self):
        return Reservoirs(copy.deepcopy(self.res_dict),phase_diagram=self.pd,
                          are_chempots_delta=self.are_chempots_delta,mu_refs=self.mu_refs)
    
    def update(self, other):
        if isinstance(other, dict):
            for key, value in other.items():
                self.res_dict[key] = value
        else:
            for key, value in other:
                self.res_dict[key] = value


    def as_dict(self):
        """
        Json-serializable dict representation of a Reservoirs object. 
        
        Returns
        -------
        dict
            Json-serializable dict of a Reservoirs object.
        """
        d = {}
        d['@module'] = self.__class__.__module__
        d['@class'] = self.__class__.__name__
        d['res_dict'] = {r:mu.as_dict() for r,mu in self.res_dict.items()} 
        d['phase_diagram'] = self.pd.as_dict() if self.pd else None
        d['mu_refs'] = self.mu_refs.as_dict() if self.mu_refs else None
        d['are_chempots_delta'] = self.are_chempots_delta
        return d


    def to_json(self,path,cls=MontyEncoder):
        """
        Save Reservoirs object as json string or file

        Parameters
        ----------
        path : (str), optional
            Path to the destination file.  If None a string is exported.
        cls : (cls)
            Encoder class for json.dump. The default is MontyEncoder.

        Returns
        -------
        d : (str)
            If path is not set a string is returned.
        """
        d = self.as_dict()
        if path:
            with open(path,'w') as file:
                json.dump(d,file,cls=cls)
            return
        else:
            return d.__str__()  


    @classmethod
    def from_dict(cls,d):
        """
        Constructor of Reservoirs object from dictionary representation.
        
        Parameters
        ----------
        d : dict
        
        Returns
        -------
        Reservoirs object.
        """
        res_dict = {}
        for res,chempots in d['res_dict'].items():
            res_dict[res] = Chempots.from_dict(chempots)
        phase_diagram = PhaseDiagram.from_dict(d['phase_diagram']) if d['phase_diagram'] else None
        mu_refs = Chempots.from_dict(d['mu_refs']) if d['mu_refs'] else None
        are_chempots_delta = d['are_chempots_delta']
            
        return cls(res_dict,phase_diagram,mu_refs,are_chempots_delta)


    @staticmethod
    def from_json(path_or_string):
        """
        Build Reservoirs object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        Reservoir object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return Reservoirs.from_dict(d)


    def filter_reservoirs(self,inplace=False,elements=None):
        """
        Get new Reservoir object filtering the chempots dictionary.

        Parameters
        ----------
        inplace : (bool)
            Apply changes to current Reservoirs object.
        elements : (list), optional
            List of element symbols. The default is None.

        Returns
        -------
        res : 
            Reservoirs object.
        """
        res = self.copy()
        mu_refs = self.mu_refs.copy()
        filtered_dict = res.res_dict
        if elements:
            d = filtered_dict.copy()
            for r in list(d):
                for el in list(d[r]):
                    if el not in elements:
                        del filtered_dict[r][el]
                        if el in mu_refs.keys():
                            del mu_refs[el]
        if inplace:
            self.res_dict = filtered_dict
            self.mu_refs = mu_refs
            return
        else:
            res.mu_refs = mu_refs
            return res


    def get_absolute_res_dict(self):
        """ 
        Return values of chempots from referenced to absolute 
        """
        if self.are_chempots_delta:
            return {r:mu.get_absolute(self.mu_refs) for r,mu in self.res_dict.items()}
        else:
            raise ValueError('Chemical potential values are already absolute')
                

    def get_referenced_res_dict(self):
        """ 
        Return values of chempots from absolute to referenced 
        """
        if self.are_chempots_delta:
            raise ValueError('Chemical potential values are already with respect to reference')
        else:
            return {r:mu.get_referenced(self.mu_refs) for r,mu in self.res_dict.items()}

    
    def get_dataframe(self,format_symbols=False,format_compositions=False,all_math=False,ndecimals=None):
        """
        Get DataFrame object of the dictionary of reservoirs

        Parameters
        ----------
        format_symbols : (bool), optional
            Format labels of element chempots as \Delta \mu_{\text{"el"}}. The default is False.
        format_compositions : (bool), optional
            Get Latex format of compositions. The default is False.
        all_math : (bool), optional
            Get all characters in composition written in Latex's math format. The default is False.
        ndecimals : (int), optional
            Number of decimals to round the chemical potentials, if None the numbers are not changed.
            The default is None.

        Returns
        -------
        df : 
            DataFrame object.
        """
        res = self._get_res_dict_with_symbols(format_symbols)
        df = DataFrame(res)
        df = df.transpose()
        if format_compositions:
            new_index = []
            for string in df.index:
                new_string = format_composition(string,all_math=all_math)
                new_index.append(new_string)
            df.index = new_index
        if ndecimals:
            df = df.round(decimals=ndecimals)
        return df
        
    
    def get_latex_table(self,ndecimals=1):
        """
        Get string with formatted latex table of chemical potentials
        """
        df = self.get_dataframe(format_symbols=True,ndecimals=ndecimals)
        table = df.to_latex(escape=False)
        return table
    
    
    def get_plot(self,elements,size=1,**kwargs):
        """
        Plot the stability diagram with the reservoir points

        Parameters
        ----------
        elements : (list)
            List of strings with element symbols on the diagram axis.
        size : (float), optional
            Size of the points. The default is 1.
        **kwargs : (dict)
            Kwargs for the add_reservoirs function.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        from pynter.phase_diagram.plotter import PDPlotterAdder # import here to avoid circular import
        res = self.copy()
        if not res.pd:
            raise ValueError('PhaseDiagram object needs to be stored to plot the stability diagram')
        plt = PDHandler(res.pd).get_stability_diagram(elements)
        if not res.are_chempots_delta:
            res.set_to_referenced()
            
        plt = PDPlotterAdder(res.pd,size).add_reservoirs(res,elements,**kwargs)
        return plt
                
    
    def set_to_absolute(self):
        """
        Set reservoir dictionary to absolute values of chempots.
        """
        new_res_dict = self.get_absolute_res_dict()
        self.res_dict = new_res_dict
        self.are_chempots_delta = False
        return

        
    def set_to_referenced(self):
        """
        Set reservoir dictionary to referenced values of chempots.
        """
        new_res_dict = self.get_referenced_res_dict()
        self.res_dict = new_res_dict
        self.are_chempots_delta = True
        return    
    
                
    def _get_res_dict_with_symbols(self,format_symbols=False):
        """
        format_labels : (bool), optional
            Format labels of element chempots as \Delta \mu_{\text{"el"}}. The default is False.
        """
        new_dict = {}
        for res,chempots in self.res_dict.items():
            new_dict[res] = {}
            chempots = chempots.mu #keep just dict for DataFrame
            for el in chempots:
                if format_symbols:
                    label = '$\Delta \mu_{\text{%s}}$' %el
                else:
                    label = el
                new_dict[res][label] = chempots[el]
        return new_dict

    
class PressureReservoirs(Reservoirs):
    """
    Subclass of Reservoirs which contains temperature information. Useful for partial pressure analysis. 
    """
    def __init__(self,res_dict,temperature=None,phase_diagram=None,mu_refs=None,are_chempots_delta=False):
        super().__init__(res_dict,phase_diagram,mu_refs,are_chempots_delta)
        self.temperature = temperature
        self.pressures = list(self.res_dict.keys())
        
    
    def __eq__(self, other):
        if isinstance(other, dict):
            return self.res_dict == other
        elif isinstance(other, PressureReservoirs):
            return self.res_dict == other.res_dict
        else:
            return False
        
    
    def as_dict(self):
        """
        Json-serializable dict representation of a PressureReservoirs object. The Pymatgen element 
        is expressed with the symbol of the element.

        Returns
        -------
        dict
            Json-serializable dict of a PressureReservoirs object.
        """
        d = {}
        d['@module'] = self.__class__.__module__
        d['@class'] = self.__class__.__name__
        d['res_dict'] = {r:mu.as_dict() for r,mu in self.res_dict.items()}
        d['temperature'] = self.temperature
        d['phase_diagram'] = self.pd.as_dict() if self.pd else None
        d['mu_refs'] = self.mu_refs.as_dict() if self.mu_refs else None
        d['are_chempots_delta'] = self.are_chempots_delta
        return d


    @classmethod
    def from_dict(cls,d):
        """
        Constructor of PressureReservoirs object from dictionary representation.
        
        Parameters
        ----------
        d : dict
        
        Returns
        -------
        PressureReservoirs object.
        """
        res_dict = {float(r):Chempots.from_dict(mu) for r,mu in d['res_dict'].items()}
        temperature = d['temperature'] if 'temperature' in d.keys() else None
        phase_diagram = PhaseDiagram.from_dict(d['phase_diagram']) if d['phase_diagram'] else None
        mu_refs = Chempots.from_dict(d['mu_refs'])  if d['mu_refs'] else None
        are_chempots_delta = d['are_chempots_delta']
            
        return cls(res_dict,temperature,phase_diagram,mu_refs,are_chempots_delta)


    @staticmethod
    def from_json(path_or_string):
        """
        Build PressureReservoirs object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        PressureReservoir object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.load(path_or_string)
        return PressureReservoirs.from_dict(d)    
    
    
    
    
class PDHandler:
    
    def __init__(self,phase_diagram):
        """
        Class to generate and handle Pymatgen phase diagram more rapidly.

        Parameters
        ----------
        phase_diagram : (PhaseDiagram)
            Pymatgen PhaseDiagram object
        """
        self.pd = phase_diagram 
        self.mu_refs = self.get_chempots_reference()


    def calculate_single_chempot(self,comp,chempots_ref):
        """
        Calculate referenced chemical potential in a given composition and given
        the chemical potentials of the other elements.

        Parameters
        ----------
        comp : (Pymatgen Composition object)
            Compositions of the phase.
        chempots_ref : (Dict)
            Dictionary with element symbols as keys and respective chemical potential as value ({el:chempot}).
            The chemical potentials used here are the ones relative to the reference (delta_mu)

        Returns
        -------
        mu : (float)
            Chemical potential.
        """        
        form_energy = self.get_formation_energy_from_stable_comp(comp)
        mu = form_energy
        for el,coeff in comp.items():
            if el.symbol in chempots_ref:
                mu += -1*coeff * chempots_ref[el.symbol]
            else:
                factor = coeff
        
        return mu/factor


    def get_all_boundaries_chempots(self,comp):
        """
        Get chemical potentials at the corners of the stability ta given composition.

        Parameters
        ----------
        comp : 
            Pymatgen Composition object.

        Returns
        -------
        chempots : (Chempots)
            Chempots object.
        """
        chempots_pmg = self.pd.get_all_chempots(comp)
        for r,mu in chempots_pmg.items():
            chempots_pmg[r] = {k:v.item() for k,v in mu.items()} # convert from np.float64
        chempots = {r:Chempots.from_pmg_elements(mu) for r,mu in chempots_pmg.items()}
        return chempots

        
    def get_chempots_reference(self):
        """
        Gets elemental reference compounds and respective e.p.a with Pymatgen el_ref attribute in PhaseDiagram class.

        Returns
        -------
        Chempots object
        """
        chempots_ref = {}
        pd = self.pd
        for el in pd.el_refs:
            chempots_ref[el] = pd.el_refs[el].energy_per_atom
        chempots_ref = {k: v for k, v in sorted(chempots_ref.items(), key=lambda item: item[0])}
        return Chempots.from_pmg_elements(chempots_ref)
       

    def get_dataframe(self):
        """
        Generate pandas DataFrame with columns 'Composition, Structure, Formation energy'.
        To display a string for 'Structure' the entry needs to be a ComputedStructureEntry (see pymatgen docs).
        """
        phases = []
        for e in self.pd.stable_entries:
            d = {}
            d['Composition'] = format_composition(e.composition.reduced_formula)
            sg = SpaceGroup(e.structure.get_space_group_info()[0])
            crystal_system, sg_string = sg.crystal_system , sg.to_latex_string()
            if e.__class__.__name__ == 'ComputedStructureEntry':
                d['Structure'] = f'{crystal_system.capitalize()} ({sg_string})'
            else:
                d['Structure'] = None
            d['Formation energy p.a (eV)'] = np.around(self.pd.get_form_energy_per_atom(e),decimals=2)
            phases.append(d)
        df = DataFrame(phases)
        return df

        
    def get_entries_from_comp(self,comp):
       """
       Get a list of entries corrisponding to the target composition.
       
       Parameters
       ----------
       comp : (Pymatgen Composition object)
       Returns
       -------
       List of Pymatgen PDEntry objects
       """
       target_entries=[]
       pd = self.pd
       for e in pd.all_entries:
           if e.composition.reduced_composition == comp:
               target_entries.append(e)
       if target_entries != []:
           return target_entries
       else:
           raise ValueError('No entry has been found for target composition:%s' %comp.reduced_formula)
               

    def get_formation_energies_from_comp(self,comp):
        """
        Get dictionary of formation energies for all entries of a given reduced composition

        Parameters
        ----------
        comp : (Pymatgen Composition object)
        Returns
        -------
        from_energies : (Dict)
            Dictionary with PDEntry objects as keys and formation energies as values.
        """
        pd = self.pd
        form_energies = {}
        for e in self.get_entries_from_comp(comp):
            factor = e.composition.get_reduced_composition_and_factor()[1]
            form_energies[e] = pd.get_form_energy(e)/factor
        return form_energies
            

    def get_formation_energy_from_stable_comp(self,comp):
        """
        Get formation energy of a target stable composition

        Parameters
        ----------
        comp : (Pymatgen Composition object)

        Returns
        -------
        Formation energy (float)
        """
        pd = self.pd
        entry = self.get_stable_entry_from_comp(comp)
        factor = entry.composition.get_reduced_composition_and_factor()[1]
        return pd.get_form_energy(entry)/factor
        

    def get_phase_boundaries_chempots(self,comp,chempot_ref):
        """
        Given a composition and a fixed chemical potential, this function analises the composition of the boundary phases
        and the associated chemical potentials at the boundaries. Only works for 3 component PD.

        Parameters
        ----------
        comp : (Pymatgen Composition object)
            Composition of the phase you want to get the chemical potentials at the boundary.
        chempot_ref : (Dict)
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}).
            The chemical potential here is the referenced value. 
        Returns
        -------
        chempots : (Dict)
            Dictionary with compositions at the boundaries as keys and delta chemical potentials as value.
        """       
        chempots = {}
        comp1,comp2 = self.get_phase_boundaries_compositions(comp, chempot_ref)
        
        boundary = '-'.join([comp1.reduced_formula,comp.reduced_formula])
        chempots[boundary] = self.solve_phase_boundary_chempots(comp1, comp, chempot_ref)
        
        boundary = '-'.join([comp.reduced_formula,comp2.reduced_formula])
        chempots[boundary] = self.solve_phase_boundary_chempots(comp, comp2, chempot_ref)
        
        return chempots      
                  
    
    def get_phase_boundaries_compositions(self,comp,chempot_ref):
        """
        Get compositions of phases in boundary of stability with a target composition given a fixed chemical potential 
        on one component. Currently only works for 3-component PD (to check). 
        Used Pymatgen GrandPotentialPhaseDiagram class. The fixed chemical potential is the referenced value that is
        converted in the global value for the analysis with the GrandPotentialPhaseDiagram class.

        Parameters
        ----------
        comp : (Pymatgen Composition object)
            Target composition for which you want to get the bounday phases.
        chempot_ref : (Dict)
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}).
            The chemical potential is the referenced value

        Returns
        -------
        comp1,comp2 : (Pymatgen Composition object)
            Compositions of the boundary phases given a fixed chemical potential for one element.
        """
        chempot_ref = Chempots(chempot_ref)
        fixed_chempot = chempot_ref.get_absolute(self.mu_refs).to_pmg_elements()
        
        entries = self.pd.all_entries
        gpd = GrandPotentialPhaseDiagram(entries, fixed_chempot)
        stable_entries = gpd.stable_entries
        comp_in_stable_entries = False
        for e in stable_entries:
            if e.original_comp.reduced_composition == comp:
                comp_in_stable_entries = True
        if comp_in_stable_entries == False:
            raise ValueError('Target composition %s is not a stable entry for fixed chemical potential: %s' %(comp.reduced_formula,fixed_chempot))
        
        el = comp.elements[0]
        x_target = comp.get_wt_fraction(el)
        x_max_left = 0
        x_min_right = 1
        for e in stable_entries:
            c = e.original_comp
            if c != comp:
                x = c.get_wt_fraction(el)
                if x < x_target and x >= x_max_left:
                    x_max_left = x
                    comp1 = c
                if x > x_target and x <= x_min_right:
                    x_min_right = x
                    comp2 = c
        return comp1.reduced_composition,comp2.reduced_composition   
                  
                    
    def solve_phase_boundary_chempots(self,comp1,comp2,chempot_ref):
        """
        Given a fixed chemical potential, gets the values of the remaining two chemical potentials
        in the boundary between two phases (region where the two phases coexist). Only works for 3-component PD (to check).
        Given a phase P1 (formula AxByOz) and a phase P2 (formula AiBjOk) the chemical potentials have to satisfy the conditions:
            form_energy(P1) = x*mu(A) + y*mu(B) +z*mu(O)
            form_energy(P2) = i*mu(A) + j*mu(B) +k*mu(O)
        From these conditions the values of mu(A) and mu(B) are determined given a fixed value of mu(O).
        All of the chemical potentials used here are delta_mu, i.e. relative to the elemental phase(delta_mu(O) = mu(O) - mu_ref(O))

        Parameters
        ----------
        comp1,comp2 : (Pymatgen Composition object)
            Compositions of the two phases at the boundary.
        chempot_ref : (Dict)
            Dictionary with fixed element symbol as key and respective chemical potential as value ({el:chempot}). The chemical potential
            used here is the one relative to the reference (delta_mu)

        Returns
        -------
        chempots_boundary : (Dict)
            Dictionary of chemical potentials.
        """
        chempots_boundary ={}
        chempot_ref = Chempots(chempot_ref)
        mu_fixed = chempot_ref.to_pmg_elements()
        for el,chempot in mu_fixed.items():
            el_fixed, mu_fixed = el, chempot
        e1 = self.get_formation_energy_from_stable_comp(comp1)
        e2 = self.get_formation_energy_from_stable_comp(comp2)
        
        coeff1 = []
        coeff2 = []
        # order of variables (mu) will follow the order of self.chempots_reference which is alphabetically ordered
        mu_refs = self.mu_refs.to_pmg_elements()
        for el in mu_refs:
            if el != el_fixed:
                if el not in comp1:
                    coeff1.append(0)
                else:
                    coeff1.append(comp1[el])
                if el not in comp2:
                    coeff2.append(0)
                else:
                    coeff2.append(comp2[el])                
        a = np.array([coeff1,coeff2])
        
        const1 = e1 - comp1[el_fixed]*mu_fixed if el_fixed in comp1 else e1 
        const2 = e2 - comp2[el_fixed]*mu_fixed if el_fixed in comp2 else e2
        b = np.array([const1,const2])
        x = np.linalg.solve(a, b)
        # output will follow the order given in input
        counter = 0
        for el in mu_refs:
            if el != el_fixed:
                chempots_boundary[el] = x[counter]
                counter += 1
        chempots_boundary[el_fixed] = mu_fixed        
        
        return Chempots.from_pmg_elements(chempots_boundary) 


    def get_plot(self,**kwargs):
        """
        Get plot with Pymatgen
        """
        PDPlotter(self.pd,show_unstable=0,backend='matplotlib').get_plot(**kwargs)
        return plt


    def get_stability_diagram(self,elements,size=None):
        """
        Method to get stability diagram with 'get_chempot_range_map_plot' method in pymatgen

        Parameters
        ----------
        elements : (list)
            List with strings of the elements to be used as free variables.
        size : (tuple)
            New size in inches.

        Returns
        -------
        plt : 
            Matplotlib object
        """
        pd = self.pd
        elements = [Element(el) for el in elements]
        PDPlotter(pd).get_chempot_range_map_plot(elements)
        if size:
            fig = plt.gcf()
            fig.set_size_inches(size[0],size[1])
        
        return plt


    def get_stable_entry_from_comp(self,comp):
        """
        Get the PDEntry of the stable entry of a target composition

        Parameters
        ----------
        comp : (Pymatgen Composition object)
        Returns
        -------
        Pymatgen PDEntry object
        """
        target_entry=None
        pd = self.pd
        for e in pd.stable_entries:
            if e.composition.reduced_composition == comp:
                target_entry = e
                break
        if target_entry is not None:
            return target_entry
        else:
            raise ValueError('No stable entry has been found for target composition:%s' %comp.reduced_formula)

    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 30 14:45:20 2023

@author: villa
"""

from pynter.slurm.core import Slurm
from pynter.slurm.job_settings import JobSettings

def setup_job_script(subparsers):
    
    parser_sub = subparsers.add_parser('job-script',help='Create job script')
    parser_sub = parse_job_script_args(parser_sub)
    parser_sub.set_defaults(func=write_job_script)
    return
    

def parse_job_script_args(parser):
    shdef = JobSettings()
    parser.add_argument('--filename',help='File name (default: %(default)s)',required=False,default=shdef.filename,type=str,metavar='',dest='filename')
    parser.add_argument('--array-size',help='Size of job array',required=False,default=shdef.array_size,type=int,metavar='',dest='array_size')
    parser.add_argument('--modules',action='append',help="Modules to load (default: %(default)s)" ,required=False,default=shdef.modules,type=str,metavar='',dest='modules')
    parser.add_argument('--exe',help='Path to executable (default: %(default)s)',required=False,default=shdef.path_exe,type=str,metavar='',dest='path_exe')
    parser.add_argument('--stop-array',action='store_true',help='Add lines to stop array jobs when converged (default: %(default)s)',required=False,default=False,dest='add_stop_array')
    parser.add_argument('--automation',help='Script with automation to add',required=False,default=shdef.add_automation,type=str,metavar='',dest='add_automation')
    parser.add_argument('--header',action='append',help='Add line to header part of script',required=False,default=shdef.add_lines_header,type=str,metavar='',dest='add_lines_header')
    parser.add_argument('--body',action='append',help='Add line to body part of script',required=False,default=shdef.add_lines_body,type=str,metavar='',dest='add_lines_body')

    slurm = Slurm()
    for key in slurm.arguments:
        if key in slurm.keys():
            parser.add_argument(f'--{key}',required=False,type=str,default=slurm[key],metavar='',dest=key)
        else:
            parser.add_argument(f'--{key}',required=False,type=str,metavar='',dest=key)

    return parser 


def write_job_script(args):
    kwargs = vars(args)
    if 'func' in kwargs.keys():
        del kwargs['func']

    JobSettings(**kwargs).write_bash_file()
    return
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 12 17:55:35 2023

@author: lorenzo
"""
import inspect
import os
import os.path as op
import unittest
import numpy as np
import warnings

from pymatgen.core.structure import Structure


class PynterTest(unittest.TestCase):
    

    def assert_all_close(self,actual, desired, rtol=1e-07, atol=0, err_msg="", verbose=True):
        """
        Tests if two arrays are almost equal up to some relative or absolute tolerance.
        """
        return np.testing.assert_allclose(actual, desired, rtol=rtol, atol=atol, err_msg=err_msg, verbose=verbose)


    def assert_object_almost_equal(self, actual, desired, rtol=1e-07, atol=0, err_msg='', verbose=True):
        """
        Custom assertion function to check if two objects are almost equal.
        Supports dictionaries, numbers (int, float), tuples, lists, str, bool and NoneType.
        Uses numpy.testing.assert_almost_equal for numerical comparisons.
        """
        if type(actual) != type(desired):
            warnings.warn(f"Objects are of different types: {type(actual)}, {type(desired)}")
    
        if isinstance(actual, dict):
            assert len(actual) == len(desired), "Dictionaries have different lengths."
            for key in actual:
                assert key in desired, f"Key '{key}' not found in the second dictionary."
                self.assert_object_almost_equal(actual[key], desired[key], rtol=rtol, atol=atol, err_msg=err_msg, verbose=verbose)
    
        elif isinstance(actual, (int, float)):
            np.testing.assert_allclose(actual, desired, rtol=rtol, atol=atol, err_msg=err_msg, verbose=verbose)
    
        elif isinstance(actual, (tuple, list)):
            assert len(actual) == len(desired), "Sequences have different lengths."
            for elem1, elem2 in zip(actual, desired):
                self.assert_object_almost_equal(elem1, elem2, rtol=rtol, atol=atol, err_msg=err_msg, verbose=verbose)
                
        elif isinstance(actual, str):
              self.assert_str_content_equal(actual, desired, err_msg=err_msg, verbose=verbose)
        
        elif isinstance(actual, (bool,type(None))):
            assert actual == desired
        

        else:
            raise AssertionError(f"Unsupported object type: {type(actual)}")


    def assert_str_content_equal(self, actual, desired, err_msg="", verbose=True):
        """
        Tests if two strings are equal, ignoring things like trailing spaces, etc.
        """
        lines1 = actual.split("\n") 
        lines2 = desired.split("\n")
        if len(lines1) != len(lines2):
            return False
        failed = []
        for l1, l2 in zip(lines1, lines2):
            if l1.strip() != l2.strip():
                failed.append(f"{l1} != {l2}")
        return len(failed) == 0
    
    
    @property
    def incar_settings(self):
        incar_settings = {'ALGO': 'Normal',
         'AMIX': 0.2,
         'EDIFF': 1e-06,
         'EDIFFG': -0.05,
         'ENCUT': 550,
         'IBRION': 2,
         'ISIF': 2,
         'ISMEAR': 0,
         'ISPIN': 1,
         'KPAR': 4,
         'LCHARG': '.TRUE.',
         'LORBIT': 10,
         'LREAL': '.FALSE.',
         'LWAVE': '.TRUE.',
         'NELM': 200,
         'NSW': 0,
         'SIGMA': 0.05,
         'SYSTEM': 'Si',
         'ISYM': 2}
        return incar_settings
 
    
    @property
    def job_settings(self):
        job_settings = {
          'slurm': {'account': 'project0000',
          'error': 'err.%j',
          'job-name': 'Si-BS_PBE-el-str_3',
          'mail-user': 'test@pynter.com',
          'mem-per-cpu': 3500,
          'nodes': 1,
          'ntasks': 384,
          'output': 'out.%j',
          'partition': 'deflt',
          'time': '00:30:00'},
         'filename': 'job.sh',
         'array_size': None,
         'modules': ['intel/2019.2', 'intel/2019.3', 'intelmpi/2019.3', 'fftw/3.3.8'],
         'path_exe': '/home/vasp-5-3-3',
         'add_stop_array': True,
         'add_automation': 'automation_vasp.py --contcar --chgcar --wavecar --check-kpoints --error-check',
         'add_lines_header': None,
         'add_lines_body': None}
        return job_settings 


    @property
    def structure(self):
        structure = Structure.from_dict(
        {'@module': 'pymatgen.core.structure',
         '@class': 'Structure',
         'charge': 0,
         'lattice': {'matrix': [[3.32548851, 0.0, 1.91997169],
           [1.10849617, 3.13530064, 1.91997169],
           [0.0, 0.0, 3.83994338]],
          'pbc': (True, True, True),
          'a': 3.839943374653261,
          'b': 3.839943378813096,
          'c': 3.83994338,
          'alpha': 59.99999998977525,
          'beta': 59.99999995393976,
          'gamma': 60.00000000512866,
          'volume': 40.036809671145996},
         'sites': [{'species': [{'element': 'Si', 'occu': 1}],
           'abc': [0.875, 0.875, 0.875],
           'xyz': [3.879736595, 2.74338806, 6.719900914999999],
           'label': 'Si',
           'properties': {}},
          {'species': [{'element': 'Si', 'occu': 1}],
           'abc': [0.125, 0.125, 0.125],
           'xyz': [0.554248085, 0.39191258, 0.959985845],
           'label': 'Si',
           'properties': {}}]}
            )
        return structure
    
    @property
    def test_files_path(self):
        """
        Path with test files ("./test_files")
        """
        module_file = inspect.getmodule(self.__class__).__file__
        module_dir = module_file.strip(op.basename(module_file))
        return op.join(module_dir,'test_files')
    
    def get_testfile_path(self,filename):
        return op.join(self.test_files_path,filename)
    

    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun  1 12:06:45 2023

@author: villa
"""

def get_dict_from_line_string(dict_string,convert_to_float=True):
    """
    Convert a string in the format "key1 value1 key2 value2 ..." to a dictionary.

    Parameters
    ----------
    dict_string : (str)
        String in the format "key1 value1 key2 value2 ...".
    convert_to_float : (bool), optional
        Convert values to float. The default is True.

    Returns
    -------
    dictionary : (dict)
    """
    items = dict_string.split()
    if len(items) % 2 != 0:
        raise ValueError("Invalid dictionary string")

    dictionary = {}
    for i in range(0, len(items), 2):
        key = items[i]
        value = items[i + 1]
        dictionary[key] = float(value) if convert_to_float else value

    return dictionary


def round_floats(obj,decimals=2):
    """
    Round floats in an object (float, dict, tuple or list)

    Parameters
    ----------
    obj : 
        Object.
    decimals : (int)
        Number of decimals to round to.

    Returns
    -------
    Object with rounded floats.

    """
    if isinstance(obj, float):
        return round(obj, decimals)
    elif isinstance(obj, dict):
        return {key: round_floats(value) for key, value in obj.items()}
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return [round_floats(element) for element in obj]
    else:
        return obj



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 17 15:24:29 2021

@author: villa
"""

import matplotlib
import matplotlib.pyplot as plt
import pandas as pd

from pynter.defects.defects import format_legend_with_charge_number, get_defect_name_from_string


class ConcBarPlotter:

    def __init__(self,concentrations,format_names=True):
        """
        Class to visualize concentrations dictionaries with pd.DataFrame and 
        pd.Series (fortotal concentrations).

        Parameters
        ----------
        concentrations : (list or dict)
            Concentrations can be in different formats: list of dict for "all" and 
            "stable_charges" and dict for "total".
        format_names : (bool), optional
            Format names with latex symbols. The default is True.
        """
        # to be fixed
        conc_dict = []
        for c in concentrations:
            d = {'charge':c.charge,'conc':c.conc,'stable':c.stable}
            if format_names:
                name = c.name.symbol
            else:
                name = c.name.fullname
            d['name'] = format_legend_with_charge_number(name,c.charge)
            conc_dict.append(d)
        
        conc_total_dict = {}
        for dn,conc in concentrations.total.items():
            if format_names:
                name = dn.symbol
            else:
                name = dn.fullname
            conc_total_dict[name] = conc

        self.conc = conc_dict
        self.conc_total = conc_total_dict
        self.format = format_names
        self.df = pd.DataFrame(self.conc)
        self.series = pd.Series(self.conc_total,name='Total Concentrations')
    

    def __print__(self):
        return self.df.__print__()
    
    def __repr__(self):
        return self.df.__repr__()

    def copy(self):
        return self.df.copy()
    
    
    def plot_bar(self,conc_range=(1e13,1e40),ylim=None,total=True,ylabel_fontsize=15,**kwargs):
        """
        Bar plot of concentrations with pd.DataFrame.plot

        Parameters
        ----------
        conc_range : (tuple), optional
            Range of concentrations to include in df. The default is (1e13,1e40).
        ylim : (tuple), optional
            Limit of y-axis in plot. If None conc_range is used. The default is None.
        total : (bool), optional
            plot total concentrations. The default is True.
        ylabel_fontsize : (int), optional
            Size of the label for y-axis. The default is 15.
        **kwargs : (dict)
            Kwargs to pass to df.plot().

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        if conc_range:
            if not ylim:
                ylim = conc_range
            series = self.limit_conc_range(conc_range,reset_df=False)[1]
            df = self.limit_conc_range(conc_range,reset_df=False)[0]
        else:
            series = self.series
            df = self.df
        if not total:
            ax = df.plot(x='name',y='conc',kind='bar',ylim=ylim,logy=True,grid=True,
                          xlabel='Name , Charge',legend=None,**kwargs)
        else:
            ax = series.plot(kind='bar',logy=True,ylim=ylim,ylabel='Concentrations(cm$^{-3}$)',grid=True,**kwargs)
        ax.set_ylabel('Concentrations(cm$^{-3}$)',fontdict={'fontsize':ylabel_fontsize})
        return ax


    def limit_conc_range(self,conc_range=(1e10,1e40),reset_df=False):
        """
        Limit df to a concentration range

        Parameters
        ----------
        conc_range : (tuple), optional
            Range of concentrations to include in df. The default is (1e10,1e40).
        reset_df : (bool), optional
            Reset df attribute or return a new df. The default is False.

        Returns
        -------
        df, series : 
            DataFrame object, Series object.
        """
        if reset_df:
            self.df = self.df[self.df.conc.between(conc_range[0],conc_range[1])]
            self.series = self.series.between(conc_range[0],conc_range[1])
            return
        else:
            df = self.df.copy()
            series = self.series.copy()
            df = df[df.conc.between(conc_range[0],conc_range[1])]
            series = series[series.between(conc_range[0],conc_range[1])]
            return df , series
    
 
    
class ThermodynamicsPlotter:
    """
    Class to plot defect Thermodynamics data.
    """
    
    def __init__(self,xlim=(1e-20,1e08)):
        
        self.xlim = xlim
        

    def plot_pO2_vs_concentrations(self,thermodata,output='total',size=(12,8),fontsize=22,
                            xlim=None,ylim=None,show_unstable=True,colors=None,**kwargs):
        """
        Plot defect and carrier concentrations in a range of oxygen partial pressure.

        Parameters
        ----------
        thermodata: (ThermoData)
            ThermoData object that contains the thermodynamic data:
                partial_pressures : (list)
                    list with partial pressure values.
                defect_concentrations : (list or dict)
                    Defect concentrations in the same format as the output of DefectsAnalysis. 
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
        output : (str), optional
            Type of output for defect concentrations:
                "all": The output is the concentration of every defect entry.
                "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
                "total": The output is the sum of the concentration in every charge for each specie.
                The default is 'total'.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        fontsize : (float), optional
            Size of font for matplotlib rcParams. The default is 22.
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.
        show_unstable : (bool), optional
            Show regions where the system is unstable (at least one formation energy is negative).
        colors : (list)
            List of colors to use for plotting with matplotlib. If None the defaults are used.
        kwargs : 
            Kwargs to pass to DefectConcentrations.filter_concentrations(**kwargs).
            If provided, only the filtered concentrations will be plotted. If output
            is set to "total", only the filtered concentrations will be used to 
            compute the total concentration.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        td = thermodata
        p,dc,cc = td.partial_pressures,td.defect_concentrations,td.carrier_concentrations
        xlabel = 'Oxygen partial pressure (atm)'
        plt = self.plot_x_vs_concentrations(x=p,xlabel=xlabel,defect_concentrations=dc,
                                                 carrier_concentrations=cc,output=output,
                                                 size=size,fontsize=fontsize,xlim=xlim,ylim=ylim,
                                                 show_unstable=show_unstable,colors=colors,**kwargs)
        return plt
    

    def plot_pO2_vs_conductivity(self,partial_pressures,conductivities,new_figure=True,
                                 label=None,size=(12,8),xlim=None,ylim=None):
        """
        Plot conductivity as a function of the oxygen partial pressure.

        Parameters
        ----------
        partial_pressures : (list)
            list with partial pressure values.
        conductivities : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and conductivity list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        xlabel = 'Oxygen partial pressure (atm)'
        plt = self.plot_x_vs_conductivity(x=partial_pressures,xlabel=xlabel,
                                          conductivities=conductivities,
                                          new_figure=new_figure,label=label,size=size,
                                          xlim=xlim,ylim=ylim)
        return plt    

    
    
    def plot_pO2_vs_fermi_level(self,partial_pressures,fermi_levels,band_gap,new_figure=True,
                         label=None,size=(12,8),xlim=None,ylim=None,colors=None):
        """
        Plot Fermi level as a function of the oxygen partial pressure.

        Parameters
        ----------
        partial_pressures : (list)
            list with partial pressure values.
        fermi_levels : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and fermi level list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        band_gap : (float)
            Band gap of the bulk material.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.
        colors : (list), optional
            List with colors for Fermi level data.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        xlabel = 'Oxygen partial pressure (atm)'
        plt = self.plot_x_vs_fermi_level(x=partial_pressures,xlabel=xlabel,
                                          fermi_levels=fermi_levels,band_gap=band_gap,
                                          new_figure=new_figure,label=label,size=size,
                                          xlim=xlim,ylim=ylim,colors=None)
        return plt


    def plot_variable_species_vs_concentrations(self,thermodata,output='total',size=(12,8),fontsize=22,
                            xlim=None,ylim=None,show_unstable=True,colors=None,**kwargs):
        """
        Plot defect and carrier concentrations in a range of oxygen partial pressure.

        Parameters
        ----------
        thermodata: (ThermoData)
            ThermoData object that contains the thermodynamic data:
                variable_defect_specie : (str)
                    Name of variable defect species.
                variable_concentrations : (list)
                    List of concentrations of variable species. 
                defect_concentrations : (list or dict)
                    Defect concentrations in the same format as the output of DefectsAnalysis. 
                carrier_concentrations : (list)
                    List of tuples with intrinsic carriers concentrations (holes,electrons).
        output : (str), optional
            Type of output for defect concentrations:
                "all": The output is the concentration of every defect entry.
                "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
                "total": The output is the sum of the concentration in every charge for each specie.
                The default is 'total'.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        fontsize : (float), optional
            Size of font for matplotlib rcParams. The default is 22.
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.
        show_unstable : (bool), optional
            Show regions where the system is unstable (at least one formation energy is negative).
        colors : (list)
            List of colors to use for plotting with matplotlib. If None the defaults are used.
        kwargs : 
            Kwargs to pass to DefectConcentrations.filter_concentrations(**kwargs).
            If provided, only the filtered concentrations will be plotted. If output
            is set to "total", only the filtered concentrations will be used to 
            compute the total concentration.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        td = thermodata
        c,dc,cc = td.variable_concentrations,td.defect_concentrations,td.carrier_concentrations
        dname = self._get_variable_defect_specie_label(td.variable_defect_specie)
        xlabel = '[%s] (cm$^{-3})$' %dname
        
        plt = self.plot_x_vs_concentrations(x=c,xlabel=xlabel,defect_concentrations=dc,
                                                 carrier_concentrations=cc,output=output,
                                                 size=size,fontsize=fontsize,xlim=xlim,ylim=ylim,
                                                 show_unstable=show_unstable,colors=colors,**kwargs)
        return plt


    def plot_variable_species_vs_conductivity(self,xlabel,variable_concentrations,conductivities,
                                              new_figure=True,label=None,size=(12,8),xlim=None,ylim=None):
        """
        Plot conductivity as a function of the oxygen partial pressure.

        Parameters
        ----------
        xlabel : (str)
            Label for concentration axis (cm^-3 is added).
        variable_concentrations : (list)
            List of concentrations of variable species. 
        conductivities : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and conductivity list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        xlabel = '%s (cm$^{-3})$' %xlabel
        plt = self.plot_x_vs_conductivity(x=variable_concentrations,xlabel=xlabel,
                                          conductivities=conductivities,
                                          new_figure=new_figure,label=label,size=size,
                                          xlim=xlim,ylim=ylim)
        return plt
        

    def plot_variable_species_vs_fermi_level(self,xlabel,variable_concentrations,fermi_levels,band_gap,
                         new_figure=True,label=None,size=(12,8),xlim=None,ylim=None,colors=None):
        """
        Plot Fermi level as a function of the oxygen partial pressure.

        Parameters
        ----------
        xlabel : (str)
            Label for concentration axis (cm^-3 is added).
        variable_concentrations : (list)
            List of concentrations of variable species. 
        fermi_levels : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and fermi level list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        band_gap : (float)
            Band gap of the bulk material.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.
        colors : (list), optional
            List with colors for Fermi level data.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        xlabel = '%s (cm$^{-3})$' %xlabel
        plt = self.plot_x_vs_fermi_level(x=variable_concentrations,xlabel=xlabel,
                                          fermi_levels=fermi_levels,band_gap=band_gap,
                                          new_figure=new_figure,label=label,size=size,
                                          xlim=xlim,ylim=ylim,colors=None)
        return plt



    def plot_x_vs_concentrations(self,x,xlabel,defect_concentrations,carrier_concentrations,
                                     output='total',size=(12,8),fontsize=22,
                                     xlim=None,ylim=None,show_unstable=True,colors=None,**kwargs):
        """
        Plot defect concentrations as a function of generic data on the x-axis.        
        
        Parameters
        ----------
        x : (list)
            List with data on x-axis.
        xlabel : (str)
            Label for x-axis.
        defect_concentrations : (list)
            List of DefectConcentrations objects.
        carrier_concentrations : (list)
            List of tuples with carrier concentrations (holes,electrons).
         output : (str), optional
             Type of output for defect concentrations:
                 "all": The output is the concentration of every defect entry.
                 "stable": The output is the concentration of the stable charge for every defect at each fermi level point.
                 "total": The output is the sum of the concentration in every charge for each specie.
                 The default is 'total'.
         size : (tuple), optional
             Size of the matplotlib figure. The default is (12,8).
         fontsize : (float), optional
             Size of font for matplotlib rcParams. The default is 22.
         xlim : (tuple), optional
             Range of x-axis. The default is (1e-20,1e08).
         ylim : (tuple), optional
             Range of y-axis. The default is None.
         show_unstable : (bool), optional
             Show regions where the system is unstable (at least one formation energy is negative).
         colors : (list)
             List of colors to use for plotting with matplotlib. If None the defaults are used.
         kwargs : 
             Kwargs to pass to DefectConcentrations.filter_concentrations(**kwargs).
             If provided, only the filtered concentrations will be plotted. If output
             is set to "total", only the filtered concentrations will be used to 
             compute the total concentration.

         Returns
         -------
         plt : 
             Matplotlib object.
         """        
        matplotlib.rcParams.update({'font.size': 22})
        if output == 'all' or output == 'stable':
            plt = self._plot_conc(x=x,defect_concentrations=defect_concentrations,
                             carrier_concentrations=carrier_concentrations,
                             output=output,size=size,colors=colors,**kwargs)
        elif output == 'total':
            plt = self._plot_conc_total(x=x,defect_concentrations=defect_concentrations,
                                   carrier_concentrations=carrier_concentrations,
                                   size=size,colors=colors,**kwargs)
        else:
            raise ValueError('The options for plot output are "all", "stable" or "total".')
            
        plt.xscale('log')
        plt.yscale('log')
        if show_unstable:
            stable = self._get_unstable_bool(defect_concentrations)
            ax = plt.gca()
            plt.fill_between(x, 0, 1, where=stable, alpha=0.3, transform=ax.get_xaxis_transform(), color='red')
        xlim = xlim if xlim else self.xlim
        plt.xlim(xlim)
        if ylim:
            plt.ylim(ylim)

        plt.xlabel(xlabel)
        plt.ylabel('Concentrations (cm$^{-3})$')
        plt.legend()
        plt.grid()
        
        return plt   

        

    def plot_x_vs_conductivity(self,x,xlabel,conductivities,new_figure=True,
                               label=None,size=(12,8),xlim=None,ylim=None):
        """
        Plot conductivity as a function of the oxygen partial pressure.

        Parameters
        ----------
        x : (list)
            List with data on x-axis.
        xlabel : (str)
            Label for x-axis.
        conductivities : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and conductivity list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        matplotlib.rcParams.update({'font.size': 22})
        if new_figure:
            plt.figure(figsize=(size))
        if isinstance(conductivities,dict):
            for name,sigma in conductivities.items():
                plt.plot(x,sigma,linewidth=4,marker='s',label=name)
        else:
            sigma = conductivities
            plt.plot(x,sigma,linewidth=4,marker='s',label=label)
        plt.xscale('log')
        plt.yscale('log')
        xlim = xlim if xlim else self.xlim
        plt.xlim(xlim)
        if ylim:
            plt.ylim(ylim)
        plt.xlabel(xlabel)
        plt.ylabel('Conductivity (S/m)')
        plt.legend()
        if new_figure:
            plt.grid()
        
        return plt


    def plot_x_vs_fermi_level(self,x,xlabel,fermi_levels,band_gap,new_figure=True,
                         label=None,size=(12,8),xlim=None,ylim=None,colors=None):
        """
        Parameters
        ----------
        x : (list)
            List with data on x-axis.
        xlabel : (str)
            Label for x-axis.
        fermi_levels : (dict or list)
            If is a dict multiples lines will be plotted, with labels as keys and fermi level list
            as values. If is a list only one line is plotted with label taken from the "label" argument.
        band_gap : (float)
            Band gap of the bulk material.
        new_figure : (bool), optional
            Initialize a new matplotlib figure. The default is True.
        label : (str), optional
            Label for the data. The default is None.
        size : (tuple), optional
            Size of the matplotlib figure. The default is (12,8).
        xlim : (tuple), optional
            Range of x-axis. The default is (1e-20,1e08).
        ylim : (tuple), optional
            Range of y-axis. The default is None.
        colors : (list), optional
            List with colors for Fermi level data.

        Returns
        -------
        plt : 
            Matplotlib object.
        """
        ylim = ylim if ylim else (-0.5, band_gap+0.5)
        matplotlib.rcParams.update({'font.size': 22})
        if new_figure:
            plt.figure(figsize=(size))
        if isinstance(fermi_levels,dict):
            for name,mue in fermi_levels.items():
                clr = colors[list(fermi_levels.keys()).index(name)] if colors else None
                plt.plot(x,mue,linewidth=4,marker='s',label=name,color=clr)
        else:
            mue = fermi_levels
            clr = colors[0] if colors else None
            plt.plot(x,mue,linewidth=4,marker='s',label=label,color=clr)
        plt.xscale('log')
        xlim = xlim if xlim else self.xlim
        plt.xlim(xlim)
        if ylim:
            plt.ylim(ylim)
        plt.hlines(0, xlim[0], xlim[1],color='k')
        plt.text(xlim[1]+(xlim[1]-xlim[0]),-0.05,'VB')
        plt.text(xlim[1]+(xlim[1]-xlim[0]),band_gap-0.05,'CB')
        plt.hlines(band_gap, xlim[0], xlim[1],color='k')
        plt.axhspan(ylim[0], 0, alpha=0.2,color='k')
        plt.axhspan(band_gap, ylim[1], alpha=0.2,color='k')

        plt.xlabel(xlabel)
        plt.ylabel('Electron chemical potential (eV)')
        plt.legend()
        if new_figure:
            plt.grid()
        
        return plt
    
    
    def _get_variable_defect_specie_label(self,variable_defect_specie):
        try:
            dname = get_defect_name_from_string(variable_defect_specie)
            return dname.symbol
        except:
            return variable_defect_specie

    
    
    def _plot_conc(self,x,defect_concentrations,carrier_concentrations,output,size,colors,**kwargs):
        
        plt.figure(figsize=size)
        dc = defect_concentrations if output != 'stable' else [c.stable for c in defect_concentrations] 
        if kwargs:
            dc = [c.filter_concentrations(**kwargs) for c in dc] #filter concentrations based on kwargs
        h = [cr[0] for cr in carrier_concentrations] 
        n = [cr[1] for cr in carrier_concentrations]
        previous_charge = None
        for i in range(0,len(dc[0])):
                conc = [c[i].conc for c in dc]
                charges = [c[i].charge for c in dc]
                label_txt = dc[0][i].name.symbol
                if output == 'all':
                    label_txt = format_legend_with_charge_number(label_txt,dc[0][i].charge)
                elif output == 'stable':
                    for q in charges:
                        if q != previous_charge:
                            previous_charge = q
                            label_charge = '+' + str(q) if q > 0 else str(q)
                            index = charges.index(q)
                            plt.text(x[index],conc[index],label_charge,clip_on=True)
                color = colors[i] if colors else None
                plt.plot(x,conc,label=label_txt,linewidth=4,color=color)
        plt.plot(x,h,label='$n_{h}$',linestyle='--',color='r',linewidth=4)
        plt.plot(x,n,label='$n_{e}$',linestyle='--',color='b',linewidth=4)
        return plt
    
    
    def _plot_conc_total(self,x,defect_concentrations,carrier_concentrations,size,colors,**kwargs):
        
        dc = defect_concentrations
        if kwargs:
            dc = [c.filter_concentrations(**kwargs) for c in dc]
        plt.figure(figsize=size)
        h = [cr[0] for cr in carrier_concentrations] 
        n = [cr[1] for cr in carrier_concentrations]
        for name in dc[0].names:
            conc = [c.total[name] for c in dc]
            label_txt = name.symbol
            color = colors[dc[0].names.index(name)] if colors else None
            plt.plot(x,conc,label=label_txt,linewidth=4,color=color)
        plt.plot(x,h,label='$n_{h}$',linestyle='--',color='r',linewidth=4)
        plt.plot(x,n,label='$n_{e}$',linestyle='--',color='b',linewidth=4)
        return plt
    
    
    def _get_unstable_bool(self,defect_concentrations):
        slist = []
        for dc in defect_concentrations:
            unstable = False
            for d in dc:
                if 'stable' in vars(d).keys() and d.stable == False:
                    unstable = True
            slist.append(unstable)
        
        return slist
        
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon May 10 14:34:48 2021

@author: villa
"""


import os
import os.path as op
import warnings
import numpy as np
import json
from glob import glob

from monty.json import MontyDecoder, MontyEncoder

from pymatgen.io.vasp.inputs import VaspInput, Poscar, Incar, Kpoints, Potcar
from pymatgen.io.vasp.outputs import Vasprun, Oszicar
from pymatgen.core.structure import Structure
from pymatgen.analysis.transition_state import NEBAnalysis

from pynter.tools.utils import grep
from pynter.data.jobs import Job
from pynter.slurm.interface import HPCInterface
from pynter.slurm.job_settings import JobSettings
from pynter import SETTINGS


from pymatgen.io.vasp.inputs import UnknownPotcarWarning

warnings.filterwarnings("ignore", category=UnknownPotcarWarning)


class VaspJob(Job):
 
    
    def as_dict(self,**kwargs):        
        """
        Get VaspJob as dictionary. The Vasprun ouput object is not exported.
        
        Parameters
        ----------
        get_band_structure : (bool), optional
            Export BandStructure as dict. The default is False.
            
        Returns:
            Json-serializable dict representation of VaspJob.
        """
        kwargs = self._parse_kwargs(**kwargs)
                
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__,
             "path_relative":self.path_relative,
             "inputs": self.inputs.as_dict(),             
             "job_settings": self.job_settings.as_dict(),
             "job_script_filename":self.job_script_filename,
             "name":self.name}
        
        d["outputs"] = json.loads(json.dumps(self.outputs,cls=MontyEncoder))
        if d["outputs"] and "Vasprun" in d["outputs"].keys():
            del d["outputs"]["Vasprun"] # Vasprun has no from_dict
        
        d["is_converged"] = self.is_converged
        # BandStructureSymmLine has problems with JSON encoding
        d["band_structure"] = json.loads(json.dumps(self.band_structure,cls=MontyEncoder)) if kwargs['get_band_structure'] else None
        return d


    def to_json(self,path,**kwargs):
        """
        Save VaspJob object as json string or file

        Parameters
        ----------
        path : (str), optional
            Path to the destination file.  If None a string is exported.
        get_band_structure : (bool), optional
            Export BandStructure as dict. The default is False.

        Returns
        -------
        d : (str)
            If path is not set a string is returned.

        """
        d = self.as_dict(**kwargs)
        if path:
            with open(path,'w') as file:
                json.dump(d,file)
            return
        else:
            return d.__str__()   

    
    @staticmethod
    def from_dict(d):
        """
        Construct VaspJob object from python dictionary.
        
        Returns
        -------
        VaspJob object
        """
        #ensure compatibility with old path format
        if 'path_relative' in d.keys() and d['path_relative']:
            path = HPCInterface().localdir + d['path_relative']
        elif 'path' in d.keys():
            path = d['path']
        inputs = VaspInput.from_dict(d['inputs'])
        job_settings = JobSettings.from_dict(d['job_settings'])
        job_script_filename = d['job_script_filename']
        name = d['name']
        outputs= MontyDecoder().process_decoded(d['outputs'])
        
        vaspjob = VaspJob(path,inputs,job_settings,outputs,job_script_filename,name)
        
        vaspjob._band_structure = MontyDecoder().process_decoded(d['band_structure']) if d['band_structure'] else None
        vaspjob._is_converged = d['is_converged']
        if outputs and 'ComputedStructureEntry' in outputs.keys():
            for k,v in vaspjob.computed_entry.data.items():
                if k not in vaspjob._default_data_computed_entry:
                    setattr(vaspjob,k,v)
        
        return vaspjob
        
    
    @staticmethod
    def from_directory(path=None,job_script_filename=None,load_outputs=True,**kwargs):
        """
        Builds VaspJob object from data stored in a directory. Input files are read using Pymatgen VaspInput class.
        Output files are read usign Pymatgen Vasprun class.
        Job settings are read from the job script file.

        Parameters
        ----------
        path : (str)
            Path were job data is stored. If None the current wdir is used. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is set in the config file.
        load_outputs : (bool)
            Load job outputs. The default is True.
        kwargs : (dict)
            Arguments to pass to Vasprun parser.
        Returns
        -------
        VaspJob object.
        
        """
        path = path if path else os.getcwd()
        inputs = VaspInput.from_directory(path)
        outputs = {}
        if load_outputs:
            if op.isfile(op.join(path,'vasprun.xml')):
                try:
                    outputs['Vasprun'] = Vasprun(op.join(path,'vasprun.xml'),**kwargs)
                except:
                    warnings.warn('Reading of vasprun.xml in "%s" failed'%path)
                    outputs['Vasprun'] = None
        
        job_script_filename = job_script_filename if job_script_filename else JobSettings().filename
        job_settings = JobSettings.from_bash_file(path,filename=job_script_filename)
        
        return VaspJob(path=path,inputs=inputs,job_settings=job_settings,
                       outputs=outputs,job_script_filename=job_script_filename)


    @staticmethod
    def from_json(path_or_string):
        """
        Build VaspJob object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        VaspJob object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return VaspJob.from_dict(d)
        

    @property
    def incar(self):
        return self.inputs['INCAR']
    
    @property
    def kpoints(self):
        return self.inputs['KPOINTS']
    
    @property
    def poscar(self):
        return self.inputs['POSCAR']
    
    @property
    def potcar(self):
        return self.inputs['POTCAR']

    @property
    def vasprun(self):
        if 'Vasprun' in self.outputs.keys():
            return self.outputs['Vasprun']
        else:
            if not op.exists(op.join(self.path,'vasprun.xml')):
                warnings.warn('"vasprun.xml" file is not present in Job directory')
            return None

    @property
    def computed_entry(self):
        if 'ComputedStructureEntry' in self.outputs.keys():
            return self.outputs['ComputedStructureEntry']
        else:
            return None
        

    @property
    def band_structure(self):
        return self._band_structure

    
    @property
    def charge(self):
        """
        Charge of the system calculated as the difference between the value of "NELECT"
        in the INCAR and the number of electrons in POTCAR. If "NELECT" is not present 
        charge is set to 0.
        """
        charge = 0
        if 'NELECT' in self.incar.keys():
            nelect = self.incar['NELECT']
            val = {}
            for p in self.potcar:
                val[p.element] = p.nelectrons
            neutral = sum([ val[el.symbol]*coeff 
                           for el,coeff in self.initial_structure.composition.items()])
            charge = neutral - nelect
        if not isinstance(charge,int):
            charge = np.around(charge,decimals=1)
        return charge


    @property
    def complete_dos(self):
        if 'complete_dos' in self.computed_entry.data.keys():
            return self.computed_entry.data['complete_dos']
        else:
            return ValueError('CompleteDos not present in ComputedStructureEntry data')
            

    @property
    def energy_gap(self):
        """Energy gap read from vasprun.xml with Pymatgen"""
        band_gap = None
        if self.computed_entry:
            band_gap = self.computed_entry.data['eigenvalue_band_properties'][0]
            
        return band_gap
 
    @property
    def cbm(self):
        """"Conduction band minimum"""
        return self.computed_entry.data['eigenvalue_band_properties'][1]  

    @property
    def vbm(self):
        """"valence band maximum"""
        return self.computed_entry.data['eigenvalue_band_properties'][2]    

    @property
    def final_energy(self):
        """Final total energy of the calculation read from vasprun.xml with Pymatgen"""
        final_energy = None
        if self.computed_entry:
            final_energy = self.computed_entry.data['final_energy']
            
        return final_energy
    
    @property
    def final_energy_per_atom(self):
        if self.final_energy:
            return self.final_energy/len(self.final_structure)
    
    
    @property
    def final_structure(self):
        """Final structure read from "vasprun.xml" with Pymatgen"""
        final_structure = None
        if self.computed_entry:
            if self.computed_entry.data['structures']:
                final_structure = self.computed_entry.data['structures'][-1]
            
        return final_structure 
    
            
    @property
    def formula(self):
        """Complete formula from initial structure (read with Pymatgen)"""
        if self.initial_structure:
            return self.initial_structure.composition.formula
        else:
            return None

    
    @property
    def hubbards(self):
        """
        Generate dictionary with U paramenters from LDAUU tag in INCAR file

        Returns
        -------
        U_dict : (dict)
            Dictionary with Elements as keys and U parameters as values.
        """
        U_dict = {}
        incar = self.incar
        if 'LDAUU' in incar.keys():
            ldauu = incar['LDAUU']
            elements = self.initial_structure.composition.elements
            if isinstance(ldauu,str):
                ldauu = ldauu.split()
            for i in range(0,len(ldauu)):
                U_dict[elements[i]] = int(ldauu[i])
        else:
            print('No LDAUU tag present in INCAR in Job "%s"' %self.name)
            
        return U_dict
        
    @property
    def initial_structure(self):
        """Initial structure read from poscar"""
        if self.poscar:
            poscar = self.poscar
            return poscar.structure            
        else:
            warnings.warn('inputs["POSCAR"] is not defined')
            return None
    
    
    @property
    def is_converged(self):
        """
        Reads Pymatgen Vasprun object and returns "True" if the calculation is converged,
        "False" if reading failed, and "None" if is not present in the outputs dictionary.
        """
        if hasattr(self,'_is_converged'):
            return self._is_converged
        else:
            return None
            

    @property
    def is_hybrid(self):
        """Returns True when "LHFCALC" is present and equal to True in incar""" 
        is_hybrid = False
        incar = self.incar
        if 'LHFCALC' in incar.keys():
            if incar['LHFCALC'] is True:
                is_hybrid = True
        
        return is_hybrid

    @property
    def is_hubbard(self):
        """Returns True if "LDAUU" tag is present in incar"""
        is_hubbard = False
        incar = self.incar
        if 'LDAUU' in incar.keys():
            is_hubbard = True
        
        return is_hubbard


    @property
    def nelectrons(self):
        """
        Number of electrons in the system. If 'NELECT' tag is in INCAR that value is returned.
        Else the sum of valence electrons from POTCAR is returned.
        """
        if 'NELECT' in self.incar.keys():
            nelect = self.incar['NELECT']
        else:
            val = {}
            for p in self.potcar:
                val[p.element] = p.nelectrons
            nelect = sum([ val[el.symbol]*self.initial_structure.composition[el] 
                           for el in self.initial_structure.composition])
        return nelect    
    
    
    @property
    def stress(self):
        """Residual stress of last ionic step"""
        if 'ionic_steps' in self.computed_entry.data.keys():
            return self.computed_entry.data['ionic_steps'][-1]['stress']
        else:
            print('Ionic steps data not present in ComputedStructureEntry')
            return None



    def delete_output_files(self,safety=True):
        """
        Delete files that aren't input files (INCAR,KPOINTS,POSCAR,POTCAR)
        """
        if safety:
            inp = input('Are you sure you want to delete outputs of Job %s ?: (y/n)' %self.name)
            if inp in ('y','Y'):
                delete = True
            else:
                delete = False
        else:
            delete= True
            
        if delete:                
            files = [f for f in os.listdir(self.path) if os.path.isfile(os.path.join(self.path, f))]
            for f in files:
                if f not in ['INCAR','KPOINTS','POSCAR','POTCAR',self.job_script_filename]:
                    os.remove(os.path.join(self.path,f))
                    print('Deleted file %s'%os.path.join(self.path,f))   
        return
                    

    def get_inputs(self,sync=False):
        """
        Read VaspInput from directory
        """
        if sync:
            self.sync_from_hpc()
        inputs = VaspInput.from_directory(self.path)
        self.inputs = inputs
        return
    
    
    def get_outputs(self,sync=False,get_output_properties=True,**kwargs):
        """
        Get outputs dictionary from the data stored in the job directory. "vasprun.xml" is 
        read with Pymatgen

        Parameters
        ----------
        sync : (float), optional
            Sync data from hpc. The default is False.
        get_output_properties : (float), optional
            Parse output properties from Vasprun. The default is True.
        **kwargs : (dict)
            Arguments for the Vasprun class in pymatgen.
        """
        if sync:
            self.sync_from_hpc()
        path = self.path
        outputs = {}
        if op.isfile(op.join(path,'vasprun.xml')):
            try:
                outputs['Vasprun'] = Vasprun(op.join(path,'vasprun.xml'),**kwargs)
            except:
                warnings.warn('Reading of vasprun.xml in "%s" failed'%path)
                outputs['Vasprun'] = None
        self.outputs = outputs
        if get_output_properties:
            self.get_output_properties()
        return

    
    def get_output_properties(self,**kwargs):
        """
        Parse outputs properties from VaspJob.outputs.

        Parameters
        ----------
        get_band_structure : (bool or dict), optional
            Get BandStructure object from vasprun. The default is False. If is a dict the BS is 
            read and the kwargs in the dict are used in the pymatgen function.
        data : (list), optional
            List of attributes of Vasprun to parse in ComputedStructureEntry.
            the attributes parsed by default are 'final_energy','structures','eigenvalue_band_properties',
            'parameters','actual_kpoints','ionic_steps'. The default is None.
        """                
        self._is_converged = self._get_convergence()
        
        self._default_data_computed_entry = SETTINGS['vasp']['computed_entry_default'] # default imports from Vasprun

        kwargs = self._parse_kwargs(**kwargs)  
        if self.vasprun:
            data = self._default_data_computed_entry 
            if kwargs['data']:
                for attr in kwargs['data']:
                    data.append(attr)
            self.outputs['ComputedStructureEntry'] = self.vasprun.get_computed_entry(data=data)
        
        if kwargs['get_band_structure']:
            if type(kwargs['get_band_structure']) == dict:
                self._band_structure = self._get_band_structure(**kwargs['get_band_structure'])
            else:
                self._band_structure = self._get_band_structure()
        else:
            None
        
        return


    def insert_in_database(self,get_doc_only=False,safety=True,check_convergence=True,**kwargs):
        """
        Get VaspJob doc and insert in pynter default database with matgendb's VaspToDbTaskDrone.

        Parameters
        ----------
        get_doc_only: (bool), optional
            Get only doc with get_task_doc but does not perform the insertion into db. Default is False.
        safety : (bool), optional
            Ask confirmation to insert job. The default is True.
        check_convergence: (bool), optional
            Insert job in DB only if is_converged is True. The default is True.
        **kwargs :
            Args to pass to VaspToDbTaskDrone
            
        Returns
        -------
        drone: 
            VaspJobDrone object that contains all attributes of VaspToDbTaskDrone.
        """
        from pynter.data.database.creator import VaspJobDrone
        
        drone = VaspJobDrone(self,**kwargs)
        if get_doc_only:
            return drone.get_task_doc_from_files()
        if safety:
            inp = input('Are you sure you want to insert VaspJob %s in database "%s", collection "%s"? (y/n) : ' 
                        %(self.name,drone.database,drone.collection))        
            if inp in ('y','Y'):
                assimilate = True
            else:
                assimilate = False
        if assimilate:
            drone.assimilate_job(check_convergence=check_convergence)
        return 

    
    def write_input(self):
        """Write "inputs" dictionary to files. The VaspInput class from Pymatgen is used."""
        self.job_settings.write_bash_file(path=self.path,filename=self.job_settings.filename)
        self.inputs.write_input(output_dir=self.path,make_dir_if_not_present=True)
        return


    def _get_band_structure(self,**kwargs):
        """Get BandStructure objects from Vasprun"""
        if self.vasprun:
            if not kwargs or 'kpoints_filename' not in kwargs.keys():
                kwargs['kpoints_filename'] = op.join(self.path,'KPOINTS')
            return self.vasprun.get_band_structure(**kwargs)
        else:
            return None
            

    def _get_convergence(self):
        """
        Reads Pymatgen Vasprun object and returns "True" if the calculation is converged,
        "False" if reading failed, and "None" if is not present in the outputs dictionary.
        """
        is_converged = None
        if self.outputs:
            if 'Vasprun' in self.outputs.keys():
                is_converged = False
                if self.vasprun:
                    vasprun = self.vasprun
                    conv_el, conv_ionic = False, False
                    if vasprun:
                        conv_el = vasprun.converged_electronic
                        conv_ionic = vasprun.converged_ionic
                    if conv_el and conv_ionic:
                        is_converged = True                    
        return is_converged


    def _parse_kwargs(self,**kwargs):
        kwargs['data'] = kwargs['data'] if 'data' in kwargs.keys() else None
        kwargs['get_band_structure'] = kwargs['get_band_structure'] if 'get_band_structure' in kwargs.keys() else False 
        return kwargs



class VaspNEBJob(Job):
    
    
    def as_dict(self):
        
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__,
            # "path": self.path, # old path version in dict 
             "path_relative":self.path_relative,           
             "job_settings": self.job_settings.as_dict(),
             "job_script_filename":self.job_script_filename,
             "name":self.name}
        
        inputs_as_dict = {}
        inputs_as_dict['structures'] = [s.as_dict() for s in self.structures]
        inputs_as_dict['INCAR'] = self.incar.as_dict()
        inputs_as_dict['KPOINTS'] = self.kpoints.as_dict()
        inputs_as_dict['POTCAR'] = self.potcar.as_dict()
        
        d["inputs"] = inputs_as_dict
        d["is_step_limit_reached"] = self.is_step_limit_reached
        d["is_converged"] = self.is_converged
        d["outputs"] = json.loads(json.dumps(self.outputs,cls=MontyEncoder))
        
        return d
    

    def to_json(self,path):
        """
        Save VaspNEBJob object as json string or file
        Parameters
        ----------
        path : (str), optional
            Path to the destination file.  If None a string is exported.
            
        Returns
        -------
        d : (str)
            If path is not set a string is returned.
        """
        d = self.as_dict()
        if path:
            with open(path,'w') as file:
                json.dump(d,file)
            return
        else:
            return d.__str__() 

    
    @staticmethod
    def from_dict(d):

        #ensure compatibility with old path format
        if 'path_relative' in d.keys() and d['path_relative']:
            path = HPCInterface().localdir + d['path_relative']
        elif 'path' in d.keys():
            path = d['path']
        inputs = {}
        inputs["structures"] = [Structure.from_dict(s) for s in d['inputs']['structures']]
        inputs["INCAR"] = Incar.from_dict(d['inputs']['INCAR'])
        inputs["KPOINTS"] = Kpoints.from_dict(d['inputs']['KPOINTS'])
        inputs["POTCAR"] = Potcar.from_dict(d['inputs']['POTCAR'])
        job_settings = JobSettings.from_dict(d['job_settings'])
        job_script_filename = d['job_script_filename']
        name = d['name']
        outputs= MontyDecoder().process_decoded(d['outputs'])
        
        vaspNEBjob = VaspNEBJob(path,inputs,job_settings,outputs,job_script_filename,name)
        
        vaspNEBjob._is_step_limit_reached = d['is_step_limit_reached']
        vaspNEBjob._is_converged = d['is_converged']
        
        return vaspNEBjob
    
    
    @staticmethod
    def from_directory(path=None,job_script_filename=None,load_outputs=True):
        """
        Builds VaspNEBjob object from data stored in a directory. Inputs dict is constructed
        by reading with Pymatgen INCAR, KPOINTS and POTCAR and creating a series of Structure 
        objects read from POSCARs in the images folders. 
        Inputs is thus a dict with "structures", "INCAR","KPOINTS","POTCAR" as keys.
        Output files are read usign Pymatgen NEBAnalysis and Vasprun classes.
        Job settings are read from the job script file.

        Parameters
        ----------
        path : (str)
            Path were job data is stored. If None the current wdir is used. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is set in the config file.
        load_outputs : (bool)
            Load job outputs. The default is True.

        Returns
        -------
        VaspNEBJob object.
        
        """                
        inputs = {}
        structures = []
        path = path if path else os.getcwd()
        path = op.abspath(path)
        dirs = [d[0] for d in os.walk(path)]
        for d in dirs:
            image_name = op.relpath(d,start=path)
            if all(c.isdigit() for c in list(image_name)): #check if folder is image (all characters in folder rel path need to be numbers)
                image_path = d
                structure = Poscar.from_file(op.join(image_path,'POSCAR')).structure
                structures.append(structure)

        inputs['structures'] = structures           
        inputs['INCAR'] = Incar.from_file(op.join(path,'INCAR'))
        inputs['KPOINTS'] = Kpoints.from_file(op.join(path,'KPOINTS'))
        inputs['POTCAR'] = Potcar.from_file(op.join(path,'POTCAR'))
        
        outputs = {}
        if load_outputs:
            try:
                outputs['NEBAnalysis'] = NEBAnalysis.from_dir(path)
            except:
                warnings.warn('NEB output reading with NEBAnalysis in "%s" failed'%path)
                outputs['NEBAnalysis'] = None
        
        job_script_filename = job_script_filename if job_script_filename else JobSettings().filename
        job_settings = JobSettings.from_bash_file(path,filename=job_script_filename)
        
        return VaspNEBJob(path=path,inputs=inputs,job_settings=job_settings,
                       outputs=outputs,job_script_filename=job_script_filename)


    @staticmethod
    def from_json(path_or_string):
        """
        Build VaspJob object from json file or string.
        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.
        Returns
        -------
        VaspJob object.
        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return VaspNEBJob.from_dict(d)

 
    def delete_outputs(self,safety=True):
        """
        Delete files that aren't input files (INCAR,KPOINTS,POSCAR,POTCAR)
        """
        if safety:
            inp = input('Are you sure you want to delete outputs of Job %s ?: (y/n)' %self.name)
            if inp in ('y','Y'):
                delete = True
            else:
                delete = False
        else:
            delete= True
            
        if delete:
            dirs = self.image_dirs
            dirs.append(self.path)
            for d in dirs:                
                files = [f for f in os.listdir(d) if os.path.isfile(os.path.join(d, f))]
                for f in files:
                    if f not in ['INCAR','KPOINTS','POSCAR','POTCAR',self.job_script_filename]:
                        os.remove(os.path.join(d,f))
                        print('Deleted file %s'%os.path.join(d,f))   
        return

    
    @property
    def images(self):
        return len(self.inputs['structures'])-2
    
    @property
    def image_dirs(self):
        """
        Directories of images for NEB calculations. Directories are selected if all characters in the
        directory name are digits.
        """
        dirs = []
        path = self.path
        path = op.abspath(path)
        for d in os.walk(path):
            directory = d[0]
            image_name = op.relpath(directory,start=path)
            if all(c.isdigit() for c in list(image_name)): #check if folder is image (all characters in folder rel path need to be numbers)
                dirs.append(directory)
        dirs.sort()
        return dirs
    
    @property
    def structures(self):
        return self.inputs['structures']

    @property
    def incar(self):
        return self.inputs['INCAR']
    
    @property
    def kpoints(self):
        return self.inputs['KPOINTS']
    
    @property
    def potcar(self):
        return self.inputs['POTCAR']


    @property
    def charge(self):
        """
        Charge of the system calculated as the difference between the value of "NELECT"
        in the INCAR and the number of electrons in POTCAR. If "NELECT" is not present 
        charge is set to 0.
        """
        charge = 0
        if 'NELECT' in self.incar.keys():
            nelect = self.incar['NELECT']
            val = {}
            for p in self.potcar:
                val[p.element] = p.nelectrons
            neutral = sum([ val[el.symbol]*self.initial_structure.composition[el] 
                           for el in self.initial_structure.composition])
            charge = neutral - nelect
        if not isinstance(charge,int):
            charge = np.around(charge,decimals=1)
        return charge


    @property
    def final_structures(self):
        """
        Final structures, taken from CONTCAR for all points aside from end-points (as done by NEBAnalysis).
        """
        return self.neb_analysis.structures


    @property
    def formula(self):
        """Complete formula from initial structure (read with Pymatgen)"""
        if self.initial_structure:
            return self.initial_structure.composition.formula
        else:
            return None

        
    @property
    def initial_structure(self):
        """Initial structure read from first element of ""structures" attribute. """
        return self.structures[0]

    
    @property
    def is_converged(self):
        """
        Reads Pymatgen Vasprun object and returns "True" if the calculation is converged,
        or the ionic step limit has been reached reading from the OSZICAR file.
        "False" if reading failed, and "None" if is not present in the outputs dictionary.
        """
        if hasattr(self,'_is_converged'):
            return self._is_converged
        else:
            return None


    @property
    def is_required_accuracy_reached(self):
        """
        True if "reached required accuracy - stopping structural energy minimisation" 
        is found in most recent out.* file. 
        False if file exists but no line is found.
        None if no out.* file exists.
        """
        return self._get_output_related_attribute('is_required_accuracy_reached')
    
    
    @property
    def is_step_limit_reached(self):
        """
        Reads number of ionic steps from the OSZICAR file with Pymatgen and returns True if 
        is equal to the step limit in INCAR file (NSW tag)
        """
        return self._get_output_related_attribute('is_step_limit_reached')


    @property
    def neb_analysis(self):
        """
        Pymatgen NEBAnalysis object read from self.outputs.
        """
        return self.outputs['NEBAnalysis']


    @property
    def nelectrons(self):
        """
        Number of electrons in the system. If 'NELECT' tag is in INCAR that value is returned.
        Else the sum of valence electrons from POTCAR is returned.
        """
        if 'NELECT' in self.incar.keys():
            nelect = self.incar['NELECT']
        else:
            val = {}
            for p in self.potcar:
                val[p.element] = p.nelectrons
            nelect = sum([ val[el.symbol]*self.initial_structure.composition[el] 
                           for el in self.initial_structure.composition])
        return nelect


    def get_inputs(self,sync=False):
        """
        Read inputs from Job directory
        """
        if sync:
            self.sync_from_hpc()
        inputs = {}
        structures = []
        path = op.abspath(self.path)
        dirs = [d[0] for d in os.walk(path)]
        for d in dirs:
            image_name = op.relpath(d,start=path)
            if all(c.isdigit() for c in list(image_name)): #check if folder is image (all characters in folder rel path need to be numbers)
                image_path = d
                structure = Poscar.from_file(op.join(image_path,'POSCAR')).structure
                structures.append(structure)

        inputs['structures'] = structures           
        inputs['INCAR'] = Incar.from_file(op.join(path,'INCAR'))
        inputs['KPOINTS'] = Kpoints.from_file(op.join(path,'KPOINTS'))
        inputs['POTCAR'] = Potcar.from_file(op.join(path,'POTCAR'))
        
        self.inputs = inputs
        return


    def get_outputs(self,sync=False,get_output_properties=True):
        """
        Read outputs from Job directory
        """
        if sync:
            self.sync_from_hpc()
        outputs = {}
        path = self.path  
        try:
            outputs['NEBAnalysis'] = NEBAnalysis.from_dir(path)
        except:
            warnings.warn('NEB output reading with NEBAnalysis in "%s" failed'%path)
            outputs['NEBAnalysis'] = None
            
        self.outputs = outputs
        if get_output_properties:
            self.get_output_properties()
        return
    

    def get_output_properties(self):
        """
        Parse outputs properties from VaspNEBJob.outputs.
        """
        
        self._is_required_accuracy_reached = self._get_ionic_relaxation_from_outfile()
        self._is_step_limit_reached = self._get_step_limit_reached()                
        self._is_converged = self._get_convergence()

        return

    
    def write_input(self,write_structures=True):
        """
        Write input files in all image directories
        """
        path = op.abspath(self.path)
        
        self.job_settings['nodes'] = self.images               
        incar = self.inputs['INCAR']
        kpoints = self.inputs['KPOINTS']
        potcar = self.inputs['POTCAR']
        job_settings = self.job_settings

        if write_structures:
            self.write_structures()
        
        incar.write_file(op.join(path,'INCAR'))
        kpoints.write_file(op.join(path,'KPOINTS'))
        potcar.write_file(op.join(path,'POTCAR'))
        job_settings.write_bash_file(path=path,filename=job_settings.filename)
        
        return

    
    def write_structures(self):
        """
        Writes POSCAR files in image directories
        """
        path = self.path
        structures = self.inputs['structures']
        for s in structures:
            index = structures.index(s)
            image_path = op.join(path,str(index).zfill(2)) #folders will be named 00,01,..,XX
            if not op.exists(image_path):
                os.makedirs(image_path)
            Poscar(s).write_file(op.join(image_path,'POSCAR'))
        return

    
    def _get_convergence(self):
        """
        Returns True if:
            - "reached required accuracy - stopping structural energy minimisation" is found in the most recent out file.
                OR
            - Ionic step limit has been reached, which means the step # in OSZICAR file matches the "NSW" tag in INCAR.
        Returns False if files named out.* exist but the no "reached required accuracy" has been found AND step limit 
            has not been reached.
        Returns None if no out.* files have been found.
        """
        is_converged = self._get_ionic_relaxation_from_outfile()
        if not is_converged:
            if self.is_step_limit_reached:
                is_converged = True
            
        return is_converged     


    def _get_output_related_attribute(self,attr):
        """Checks if attribute has been defined and gives value, if not defined returns None."""
        attr = '_' + attr
        if hasattr(self, attr):
            return getattr(self, attr)
        else:
            return None


    def _get_ionic_relaxation_from_outfile(self):
        """
        Useful for NEB because Pymatgen fails to read vasprun file for NEB calculations.
        This function reads the outfile with highest number in the dir and checks for the 
        string: "reached required accuracy - stopping structural energy minimisation". 
        """
        reached_accuracy = None
        outfiles = glob(os.path.join(self.path,'out*'))
        if outfiles:
            outfiles.sort()
            outfile = outfiles[-1] #taking more recent out file ("out.jobid" with bigger job id)
            lines = grep('reached required accuracy - stopping structural energy minimisation',outfile)
            if lines:
                print('"reached required accuracy - stopping structural energy minimisation" found in %s' %outfile)
                reached_accuracy = True
            else:
                reached_accuracy = False
        
        return reached_accuracy
        
        
    def _get_step_limit_reached(self):
        """
        Reads number of ionic steps from the OSZICAR file with Pymatgen and returns True if 
        is equal to the step limit in INCAR file (NSW tag)
        """
        limit_reached = True
        image_dirs = self.image_dirs
        for d in image_dirs:
            if d != image_dirs[0] and d != image_dirs[-1]:
                if not os.path.isfile(os.path.join(d,'OSZICAR')): # check if OSZICAR files are present 
                    limit_reached = False
                else:                
                    n_steps = len(Oszicar(os.path.join(d,'OSZICAR')).ionic_steps)
                    nsw = Incar.from_file(op.join(self.path,'INCAR'))['NSW'] # check NSW from INCAR in parent directory
                    if nsw != n_steps:
                        limit_reached = False
        return limit_reached    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Apr  4 15:26:50 2023

@author: villa
"""
from pymatgen.io.vasp.outputs import Vasprun, Locpot, VolumetricData, Outcar
from pynter.defects.pmg.pmg_defects_core import Vacancy , DefectEntry, Interstitial, Substitution
from pynter.defects.pmg.pmg_defects_corrections import FreysoldtCorrection, KumagaiCorrection
import os.path as op
import importlib
from pymatgen.core.sites import PeriodicSite
from pynter.tools.structure import is_site_in_structure_coords
from pynter.defects.structure import defect_finder


def get_freysoldt_correction(defect_type, defect_specie, path_to_defect_locpot,path_to_pure_locpot,charge,
                             dielectric_constant,defect_site_coordinates,energy_cutoff=500,get_plot=False):
    
    ''' Function to perform charge corrections according to the method proposed py Freysoldt
        If this correction is used, please reference Freysoldt's original paper.
        doi: 10.1103/PhysRevLett.102.016402
        
        Args:
            defect_type: 'Vacancy', 'Interstitial' or 'Substitution'
            defect_specie: string with element occupying the defect site
            path_to_defect_locpot: path to LOCPOT file of defect structure
            path_to_pure_locpot: path to LOCPOT file of Pure structure
            charge: Charge of the defected system
            dielectric_constant: Dielectric constant
            defect_site_coordinates: numpy array with fractional coordinates of defect site
            energy_cutoff: Cut-off of plane wave expansion
            get_plot: return also Matplotlib object with plot
            
        Returns:
            Freysoldt corrections values as a dictionary 
            '''
    # acquiring data from LOCPOT files    
    locpot_pure = Locpot.from_file(path_to_pure_locpot)
    vol_data_pure = VolumetricData(locpot_pure.structure,locpot_pure.data)
    
    locpot_defect = Locpot.from_file(path_to_defect_locpot)
    vol_data_defect = VolumetricData(locpot_defect.structure,locpot_defect.data)
    
    parameters = {}
    parameters['axis_grid'] = []
    parameters['bulk_planar_averages'] = []
    parameters['defect_planar_averages'] = []
    for i in range(0,3):
        parameters['axis_grid'].append(vol_data_pure.get_axis_grid(i))
        parameters['bulk_planar_averages'].append(vol_data_pure.get_average_along_axis(i))
        parameters['defect_planar_averages'].append(vol_data_defect.get_average_along_axis(i))
    parameters['initial_defect_structure'] = locpot_defect.structure
    parameters['defect_frac_sc_coords'] = defect_site_coordinates
    
    structure_bulk = locpot_pure.structure
    defect_site = PeriodicSite(defect_specie, coords=defect_site_coordinates, lattice = locpot_pure.structure.lattice)
    
    module = importlib.import_module("pynter.defects.pmg.pmg_defects_core")
    defect_class = getattr(module,defect_type)
    defect = defect_class(structure_bulk, defect_site, charge=charge, multiplicity=None)
    defect_entry = DefectEntry(defect,None,corrections=None,parameters=parameters)
    
    freysoldt_class = FreysoldtCorrection(dielectric_constant,energy_cutoff=energy_cutoff)
    
    freysoldt_corrections = freysoldt_class.get_correction(defect_entry)
  
    if get_plot:
        plt = freysoldt_class.plot(1)
        return freysoldt_corrections , plt
    else:    
        return freysoldt_corrections


def get_freysoldt_correction_from_jobs(job_defect,job_bulk,dielectric_constant,energy_cutoff=500,tol=1e-03,get_plot=False):
    """
    Get Freysoldt corrections from VaspJob objects.

    Parameters
    ----------
    job_defect : (VaspJob)
        Defect calculation.
    job_bulk : (VaspJob)
        Bulk calculation.
    dielectric_constant : (float)
        Dielectric constant
    energy_cutoff : (int)
        Cut-off of plane wave expansion in eV
    tol : (float)
        Tolerance for defect_finder function. The default is 1e-03.
    get_plot : (bool), optional
        Get Matplotlib object with plot. The default is False.

    Returns
    -------
    corr : (dict or tuple)
        Dictionary with corrections, if get_plot is True a tuple with dict and plt object is returned.

    """
    structure_defect = job_defect.initial_structure
    structure_bulk = job_bulk.final_structure
    path_to_defect_locpot = op.join(job_defect.path,'LOCPOT')
    path_to_pure_locpot = op.join(job_bulk.path,'LOCPOT')

    df_found = defect_finder(structure_defect, structure_bulk, tol=tol)
    defect_site, defect_type = df_found.site, df_found.defect_type
    defect_specie = defect_site.specie.symbol
    defect_site_coordinates = defect_site.frac_coords
    charge = job_defect.charge

    corr = get_freysoldt_correction(defect_type, defect_specie, path_to_defect_locpot, path_to_pure_locpot, 
                                    charge, dielectric_constant, defect_site_coordinates,energy_cutoff,get_plot)
    
    return corr



def get_kumagai_correction(structure_defect,structure_bulk,path_to_defect_outcar,path_to_bulk_outcar,dielectric_tensor,
                           charge,defect_type=None,defect_specie=None,defect_site=None,sampling_radius=None,gamma=None,
                           tol=1e-03,get_plot=False):
    """
    Get Kumagai correction with Pymatgen.

    Parameters
    ----------
    structure_defect : (Structure)
        Structure of defect.
    structure_bulk : (Structure)
        Bulk structure.
    path_to_defect_outcar : (str)
        Path to OUTCAR of defect calculation.
    path_to_bulk_outcar : (str)
        Path to OUTCAR of pure calculation.
    dielectric_tensor : (array or float)
        Dielectric tensor, if is a float a diagonal matrix is constructed.
    charge : (int or float)
        Charge of the defect.
    defect_type : (str), optional
        Type of defect ('Vacancy','Interstitial' or 'Substitution')
        If None it's determined with defect_finder. The default is None.
    defect_specie : (str), optional
        Symbol of the defect specie.
        If None it's determined with defect_finder. The default is None.
    defect_site : (Site), optional
        Site of defect. If None it's determined with defect_finder. The default is None.
    sampling_radius (float): radius (in Angstrom) which sites must be outside
        of to be included in the correction. Publication by Kumagai advises to
        use Wigner-Seitz radius of defect supercell, so this is default value.
    gamma (float): convergence parameter for gamma function.
                    Code will automatically determine this if set to None.
    tol : (float)
        Tolerance for comparing sites and defect_finder function. The default is 1e-03.
    get_plot : (bool), optional
        Get Matplotlib object with plot. The default is False.

    Returns
    -------
    corr : (dict or tuple)
        Dictionary with corrections, if get_plot is True a tuple with dict and plt object is returned.
    """
    
    if not defect_site and not defect_type and not defect_specie:
        df_found = defect_finder(structure_defect, structure_bulk, tol=tol)
        defect_site, defect_type = df_found.site, df_found.defect_type
        defect_specie = defect_site.specie.symbol
   
    site_matching_indices = []
    for site in structure_defect:
        site_in_str ,index_bulk = is_site_in_structure_coords(site, structure_bulk,tol=tol)
        if site_in_str:
            site_matching_indices.append([index_bulk,structure_defect.index(site)])
        else:
            print(f'Warning in Kumagai corrections: Site {site} is not in bulk structure')
    
    bulk_atomic_site_averages = Outcar(op.join(path_to_bulk_outcar,'OUTCAR')).read_avg_core_poten()[-1]
    defect_atomic_site_averages = Outcar(op.join(path_to_defect_outcar,'OUTCAR')).read_avg_core_poten()[0]
    defect_frac_sc_coords = defect_site.frac_coords
    initial_defect_structure = structure_defect
    
    parameters = {}
    parameters['bulk_atomic_site_averages'] = bulk_atomic_site_averages
    parameters['defect_atomic_site_averages'] = defect_atomic_site_averages
    parameters['site_matching_indices'] = site_matching_indices
    parameters['initial_defect_structure'] = initial_defect_structure
    parameters['defect_frac_sc_coords'] = defect_frac_sc_coords
    
    module = importlib.import_module("pynter.defects.pmg.pmg_defects_core")
    defect_class = getattr(module,defect_type)
    defect = defect_class(structure_bulk, defect_site, charge=charge, multiplicity=1)
    defect_entry = DefectEntry(defect,None,corrections=None,parameters=parameters)

    kumagai = KumagaiCorrection(dielectric_tensor,sampling_radius,gamma)
    kumagai_corrections = kumagai.get_correction(defect_entry)
    
    if get_plot:
        plt = kumagai.plot()
        return kumagai_corrections , plt
    else:    
        return kumagai_corrections
    
    
def get_kumagai_correction_from_jobs(job_defect,job_bulk,dielectric_tensor,sampling_radius=None,
                                     gamma=None,tol=1e-03,get_plot=False):
    """
    Get Kumagai corrections from VaspJob objects.

    Parameters
    ----------
    job_defect : (VaspJob)
        Defect calculation.
    job_bulk : (VaspJob)
        Bulk calculation.
    dielectric_tensor : (array or float)
        Dielectric tensor, if is a float a diagonal matrix is constructed.
    sampling_radius (float): radius (in Angstrom) which sites must be outside
        of to be included in the correction. Publication by Kumagai advises to
        use Wigner-Seitz radius of defect supercell, so this is default value.
    gamma (float): convergence parameter for gamma function.
                    Code will automatically determine this if set to None.
    tol : (float)
        Tolerance for defect_finder function. The default is 1e-03.
    get_plot : (bool), optional
        Get Matplotlib object with plot. The default is False.

    Returns
    -------
    corr : (dict or tuple)
        Dictionary with corrections, if get_plot is True a tuple with dict and plt object is returned.

    """
    
    structure_defect = job_defect.initial_structure
    structure_bulk = job_bulk.final_structure
    path_to_defect_outcar = op.join(job_defect.path)
    path_to_bulk_outcar = op.join(job_bulk.path)

    df_found = defect_finder(structure_defect, structure_bulk, tol=tol)
    defect_site, defect_type = df_found.site, df_found.defect_type
    defect_specie = defect_site.specie.symbol
    
    charge = job_defect.charge
    
    corr = get_kumagai_correction(structure_defect, structure_bulk, path_to_defect_outcar, path_to_bulk_outcar,
                                  dielectric_tensor, charge, defect_type, defect_specie, defect_site, sampling_radius,
                                  gamma, tol, get_plot)
    
    return corr




import os
import os.path as op
import pandas as pd
import json
from monty.json import MontyDecoder

from pynter.data.jobs import get_job_from_directory
from pynter.slurm.interface import HPCInterface
from pynter.tools.utils import get_object_feature, select_objects, sort_objects

def _check_job_script(job_script_filenames,files):
    """
    Check if job script names are in a list of files. Can be either a str or list of str
    """
    check = False
    if isinstance(job_script_filenames,str):    
        if job_script_filenames in files:
            check= True
            job_script_filename = job_script_filenames
        else:
            job_script_filename = None
    elif isinstance(job_script_filenames,list):
        job_script_filename = None
        for s in job_script_filenames:
            if s in files:
                check = True
                job_script_filename = s

    else:
        raise ValueError('job_script_filenames must be a string or a list of strings')

    return check,job_script_filename


def find_jobs(path,job_script_filenames='job.sh',sort='name',load_outputs=True,jobs_kwargs=None):
    """
    Find jobs in all folders and subfolders contained in path.
    The folder containing jobs are selected based on the presence of the file job_script_filename

    Parameters
    ----------
    path : (str)
        Parent directory.
    job_script_filenames : (str or list), optional
        Filename of job bash script. The default is 'job.sh'. Can also be a list of strings if multiple 
        file names are present. The default is 'job.sh'.
    sort : (str or list), optional
        Sort list of jobs by features. If False or None jobs are not sorted. The default is 'name'.
    load_outputs : (bool)
        Load job outputs. The default is True.
    jobs_kwargs : (dict), optional
        Dictionary with job class name as keys and kwargs as values. Kwargs to be used when importing job 
        from directory for each job class.

    Returns
    -------
    jobs : (list)
        List of Job objects.

    """
    jobs = []    
    for root , dirs, files in os.walk(path):
        if files != []:
            check_job_script, job_script_filename = _check_job_script(job_script_filenames,files)
            if check_job_script:
                j = get_job_from_directory(path=root,job_script_filename=job_script_filename,
                                           load_outputs=load_outputs,jobs_kwargs=jobs_kwargs)
                jobs.append(j)
    if sort:
        jobs = Dataset().sort_jobs(jobs_to_sort=jobs,features=sort)
                
    return jobs



class Dataset:
    
    def __init__(self,jobs=None,path=None,name=None,sort='name'): 
        """
        Class to store sets of calculations

        Parameters
        ----------
        jobs : (list), optional
            List of Job objects belonging to the dataset. The default is None.        
        path : (str), optional
            Path of dataset directory. If None the work dir is used if jobs is None,
            else the commonpath between the jobs is used. The default is None.
        name : (str), optional
            Name to assign to dataset. The default is None. If None the folder name is used.
        sort : (str or list), optional
            Sort list of jobs by feature. If False or None jobs are not sorted. The default is 'name'.
        """
        if jobs:
            path = op.commonpath([j.path for j in jobs])
            self.path = op.abspath(path)
        else:
            self.path = op.abspath(path) if path else os.getcwd()
        self.name = name if name else op.basename(self.path)
        self.sort = sort
        self.jobs = jobs
        if jobs:
            self._group_jobs()
            if sort:
                self.sort_jobs(reset=True,features=sort)

        self._localdir = HPCInterface().localdir
        self._workdir = HPCInterface().workdir
        self.path_relative = self.path.replace(self._localdir,'')
        
        self.path_in_hpc = self._workdir + self.path_relative


    def __str__(self):     
        return self.jobs_table().__str__()
            
    def __repr__(self):
        return self.__str__()
    
    def __iter__(self):
        return self.jobs.__iter__()
    
    def __getitem__(self,index):
        return self.jobs.__getitem__(index)


    def as_dict(self,**kwargs):
        d = {"@module": self.__class__.__module__,
             "@class": self.__class__.__name__,
             #"path":self.path, old path version in dict
             "path_relative":self.path_relative,
             "name":self.name,
             "jobs":[j.as_dict(**kwargs) for j in self.jobs],
             "sort":self.sort}
        return d
    

    def to_json(self,path='',**kwargs):
        """
        Save Dataset object as json string or file.

        Parameters
        ----------
        path : (str), optional
            Path to the destination file. If '' the path is set to "self.path/self.name.json".
            If None a string is exported. 

        Returns
        -------
        d : (str)
            If path is not set a string is returned.

        """
        d = self.as_dict(**kwargs)
        if path == '':
            path = op.join(self.path,self.name+'.json')
        if path:
            with open(path,'w') as file:
                json.dump(d,file)
            return
        else:
            return d.__str__()   

    
    @classmethod
    def from_dict(cls,d):
        #ensure compatibility with old path format
        if 'path_relative' in d.keys() and d['path_relative']:
            path = HPCInterface().localdir + d['path_relative']
        elif 'path' in d.keys():
            path = d['path']
        name = d['name']
        jobs = []
        for j in d['jobs']:
            jobs.append(MontyDecoder().process_decoded(j))
        sort = d['sort']
        
        return cls(jobs,path,name,sort)
    
    
    @staticmethod
    def from_json(path_or_string):
        """
        Build Dataset object from json file or string.

        Parameters
        ----------
        path_or_string : (str)
            If an existing path to a file is given the object is constructed reading the json file.
            Otherwise it will be read as a string.

        Returns
        -------
        Dataset object.

        """
        if op.isfile(path_or_string):
            with open(path_or_string) as file:
                d = json.load(file)
        else:
            d = json.loads(path_or_string)
        return Dataset.from_dict(d)
        
    
    
    @staticmethod
    def from_directory(path=None,job_script_filenames='job.sh',sort='name',load_outputs=True,jobs_kwargs=None): 
        """
        Static method to build Dataset object from a directory. Jobs are selected based on where the job bash script
        is present. VaspJobs are selected based on where all input files are present (INCAR,KPOINTS,POSCAR,POTCAR).

        Parameters
        ----------
        path : (str)
            Parent directory of the dataset. If None the current wdir is used.
       job_script_filenames : (str or list), optional
            Filename of job bash script. The default is 'job.sh'. Can also be a list of strings if multiple 
            file names are present. The default is 'job.sh'.
        sort : (str or list), optional
            Sort list of jobs by feature. If False or None jobs are not sorted. The default is 'name'.
        load_outputs : (bool)
            Load job outputs. The default is True.
        jobs_kwargs : (dict), optional
            Dictionary with job class name as keys and kwargs as values. Kwargs to be used when importing job 
            from directory for each job class.
        """
        path = path if path else os.getcwd()
        jobs = find_jobs(path,job_script_filenames=job_script_filenames,sort='name',
                         load_outputs=load_outputs,jobs_kwargs=jobs_kwargs) 
        
        return  Dataset(path=path,jobs=jobs,sort=sort)
    
    
    @property
    def groups(self):
        """Directory names of the first subdirectories in the dataset path."""
        groups = []
        commonpath = op.commonpath([op.abspath(j.path) for j in self.jobs])
        grps = []
        for j in self.jobs:
            job_group = op.relpath(op.abspath(j.path),start=commonpath).split('/')[0]
            if job_group not in grps:
                grps.append(job_group)
        for g in grps:
            if list(g)[0] != '.':
                groups.append(g)
        return groups


    def _group_jobs(self):
        """
        Create "groups" and "nodes" attributes based on the common subpaths of the jobs.
        The first subdirectories of the parent dataset dir will be the groups. All of the following
        subpaths will be nodes
        """
        path = op.abspath(self.path)
        groups = self.groups
        for group in groups:
            group_path = op.join(path,group)
            for job in self.jobs:
                jpath = op.abspath(job.path)
                if f'/{group}/' in jpath or op.basename(jpath) == group:
                    nodes = jpath.replace(op.commonpath([group_path,jpath]),'')
                    job.group = group
                    job.nodes = nodes
                    job.node_points = nodes.split('/')[1:]
        return


    def add_jobs(self,jobs,regroup=True):
        """
        Add Jobs to Dataset.

        Parameters
        ----------
        jobs : (list)
            List of Job objects.
        regroup : (bool), optional
            If True Jobs in Datasets are regroupped and self.path is set to the commonpath. The default is True.
        """
        if not self.jobs:
            self.jobs = []
        for j in jobs:
            self.jobs.append(j)
        if regroup:
            paths = [j.path for j in self.jobs]
            paths.append(self.path)
            commonpath = op.commonpath(paths)
            self.regroup_jobs(path=commonpath)
        return

    
    def add_jobs_from_directory(self,path,job_script_filenames='job.sh',sort='name',load_outputs=True,regroup=True):
        """
        Add jobs to the Dataset searching all folders and subfolders contained in given path. 
        Jobs are selected based on where the job bash script is present. 
        VaspJobs are selected based on where all input files are present (INCAR,KPOINTS,POSCAR,POTCAR).

        Parameters
        ----------
        path : (str)
            Parent directory of the dataset.
       job_script_filenames : (str or list), optional
            Filename of job bash script. The default is 'job.sh'. Can also be a list of strings if multiple 
            file names are present. The default is 'job.sh'.
        sort : (str or list), optional
            Sort list of jobs by feature. If False or None jobs are not sorted. The default is 'name'.
        regroup : (bool), optional
            Regroup jobs after adding new jobs list, self.path is set to the commonpath. The default is True.
        """        
        path = op.abspath(path)
        jobs = find_jobs(path,job_script_filenames=job_script_filenames,sort=sort,load_outputs=load_outputs)
        self.add_jobs(jobs,False)
        if regroup:
            commonpath = op.commonpath([path,self.path])
            self.regroup_jobs(path=commonpath)
        return
 

    def create_job(self,cls,group='',nodes='',inputs=None,job_settings=None,
                   outputs=None,job_script_filename='job.sh',name=None):
        """
        Create Job object and add it to the dataset

        Parameters
        ----------
        cls : (str) 
            Job class.
        group : (str), optional
            Name of the group to which the job belongs. The default is ''.
        nodes : (str), optional
            Name of the nodes of the job. The default is ''.
        inputs : (dict), optional
            Dictionary with input data. The default is None.
        job_settings : (dict), optional
            Dictionary with job settings. The default is None. Documentation in SbatchScript class in slurm.job_script module
        outputs : (dict), optional
            Dictionary with output data. The default is None.
        job_script_filename : (str), optional
            Filename of job script. The default is 'job.sh'.
        name : (str)
            Name of the job. If None the name is searched in the job script.
        """
        
        path = op.join(self.path,group,nodes)        
        job = cls(path=path,inputs=inputs,job_settings=job_settings,outputs=outputs,
                       job_script_filename=job_script_filename,name=name)
        job.group = group
        job.nodes = nodes
        if self.jobs:
            self.jobs.append(job)
        else:
            self.jobs = [job]
        
        return
        

    def delete_jobs(self,jobs,delete_files=True,safety=True):
        
        for j in jobs:
            if delete_files:
                j.delete_job_files(safety=safety)
            self.jobs.remove(j)
            print('Job "%s" removed from Dataset'%j.name)
        return
        
            
    def filter_jobs(self,inplace=False,jobs=None,mode='and',exclude=False,names=None,groups=None,
                    common_group=None,common_node=None,function=None,**kwargs):
        """
        Function to filter jobs based on different selection criteria.
        The priority of the selection criterion follows the order of the input
        parameters. When more than one criterion is present, the arg "mode" 
        determines the selection criterion.

        Parameters
        ----------
        inplace : (bool), optional
            If True update the current object, otherwise returns a new Dataset object.
        jobs : (list), optional
            List of Jobs to search, if None the self.jobs is used. The default is None.
        mode : (str), optional
            Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
        exclude : (bool), optional
            Exclude the jobs satisfying the criteria instead of selecting them. The default is False.
        names : (list), optional
            List of Job names. 
        groups : (list), optional
            List of groups that jobs need to belong to.
        common_group : (str), optional
            String that needs to be present in the group.
        common_node : (str), optional
            String that needs to be present in the node.
        function : (function), optional
            Specific funtion for more complex criteria. The function needs to return a bool.
        **kwargs : (dict)
            Properties that the jobs need to satisfy. Keys are referred to attributes/methods 
            present in the relative Job class. To address more than one condition relative to
            the same attribute, use lists or tuples (e.g. charge=[0,1]).

        Returns
        -------
        dataset : (Dataset)
            Dataset object with filtered jobs. If inplace is True self.jobs is updated with
            the selected jobs.
        """
        jobs = self.select_jobs(jobs=jobs,mode=mode,exclude=exclude,names=names,groups=groups,
                                common_group=common_group,common_node=common_node,
                                function=function,**kwargs)
        
        if inplace:
            self.__init__(jobs,self.path,self.name,self.sort)
            return
        else:
            return Dataset(jobs,self.path,self.name,self.sort)
        
    
    def get_jobs_inputs(self,**kwargs):
        """Read inputs for all jobs from the data stored in the respective directories"""
        for j in self.jobs:
            j.get_inputs(**kwargs)
        return

    def get_jobs_outputs(self,update_only=False,**kwargs):
        """Read output for all jobs from the data stored in respective directories"""
        for j in self.jobs:
            if update_only:
                if not j.outputs:
                    j.get_outputs(**kwargs)
            else:
                j.get_outputs(**kwargs)
        return
            
    def get_jobs_output_properties(self,**kwargs):
        """
        Get Job output attributes by reading data in outputs dictionary
        """
        for j in self.jobs:
            j.get_output_properties(**kwargs)
        return

    
    def jobs_table(self,jobs=[],status=False,display=[]):
        """
        Create a pandas DataFrame object to display the jobs in the dataset.

        Parameters
        ----------
        jobs : (list), optional
            List of jobs to display in the table. The default is []. If [] the attribute "jobs" is used.
        status : (bool), optional
            Display job status in table.
        display : (list), optional
            List of kwargs with methods in the Job class. The properties referred to these will be added
            to the table. See get_object_feature for more details. The default is [].

        Returns
        -------
        df : (DataFrame object)

        """
        jobs = jobs if jobs else self.jobs                           
        table = []
        index = []
        if status:
            stdout,stderr = HPCInterface().qstat(printout=False)
        for j in jobs:
            index.append(j.name)
            d = {}
            d['formula'] = j.formula
            d['group'] = j.group
            d['nodes'] = j.nodes
            d['is_converged'] = j.is_converged
            if status:
                d['status'] = j.get_status_from_queue(stdout)
            for feature in display:
                if isinstance(feature,list):
                    key = feature[0]
                    for k in feature[1:]:
                        key = key + '["%s"]'%k
                else:
                    key = feature
                d[key] = get_object_feature(j,feature)
            table.append(d)
            
        df = pd.DataFrame(table,index=index)
        df.index.name = 'job_name'
        
        return df
            

    def insert_jobs_in_database(self,safety=True,check_convergence=True,**kwargs):
        """
        Insert all jobs into relative database based on their class.

        Parameters
        ----------
        safety : (bool), optional
            Ask confirmation to insert every job in database. The default is True.
        check_convergence: (bool), optional
            Insert jobs in DB only if is_converged is True. The default is True.
        **kwargs :
            Args to pass to specific Job Drone.
        """
        for j in self.jobs:
            j.insert_in_database(safety=safety,check_convergence=check_convergence,**kwargs)
        return


    def queue(self,stdouts=False):
        """
        Display queue from HPC. If stdouts is True returns out and err strings.
        
        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        hpc = HPCInterface()
        stdout,stderr = hpc.qstat()
        if stdouts:
            return stdout,stderr
        else:
            return


    def regroup_jobs(self,path=None):
        """
        Regroup jobs from a given path. If path is None the current self.path is used.
        This method is different from _group_jobs since it will reset the self.path 
        attribute to the given path.
        """
        path = path if path else op.abspath(self.path)
        self.path = path
        gjobs = {}
        groups = self.groups
        for group in groups:
            gjobs[group] = {}
            group_path = op.join(path,group)
            for job in self.jobs:
                if group in job.path:
                    jpath = op.abspath(job.path)
                    nodes = jpath.replace(op.commonpath([group_path,jpath]),'')
                    gjobs[group][nodes] = job
                    job.group = group
                    job.nodes = nodes 
                    job.node_points = nodes.split('/')[1:]
        return
            

    def select_jobs(self,jobs=None,mode='and',exclude=False,names=None,groups=None,common_group=None,
                    common_node=None,function=None,**kwargs):
        """
        Function to filter jobs based on different selection criteria.
        The priority of the selection criterion follows the order of the input
        parameters. When more than one criterion is present, the arg "mode" 
        determines the selection criterion.

        Parameters
        ----------
        jobs : (list), optional
            List of Jobs to search, if None the self.jobs is used. The default is None.
        mode : (str), optional
            Filtering mode, possibilities are: 'and' and 'or'. The default is 'and'. 
        exclude : (bool), optional
            Exclude the jobs satisfying the criteria instead of selecting them. The default is False.
        names : (list), optional
            List of Job names. 
        groups : (list), optional
            List of groups that jobs need to belong to.
        common_group : (str), optional
            String that needs to be present in the group.
        common_node : (str), optional
            String that needs to be present in the node.
        function : (function), optional
            Specific funtion for more complex criteria. The function needs to return a bool.
        **kwargs : (dict)
            Properties that the jobs need to satisfy. Keys are referred to attributes/methods 
            present in the relative Job class. To address more than one condition relative to
            the same attribute, use lists or tuples (e.g. charge=[0,1]).

        Returns
        -------
        output_jobs : (list)
            List with selected jobs.
        """
        input_jobs = jobs.copy() if jobs else self.jobs.copy() 
        functions = []
        
        if names:
            def fnames(job):
                return job.name in names
            functions.append(fnames)
        
        if groups:
            def fgroups(job):
                return job.group in groups
            functions.append(fgroups)
        
        if common_group:
            def fcommon_group(job):
                return common_group in job.group
            functions.append(fcommon_group)
        
        if common_node:
            def fcommon_node(job):
                return common_node in job.nodes
            functions.append(fcommon_node)
        
        if function:
            functions.append(function)
            
        return select_objects(objects=input_jobs,mode=mode,exclude=exclude,
                              functions=functions,**kwargs)


    def sort_jobs(self,jobs_to_sort=None,features=['name'],reset=True,reverse=False):
        """
        Sort jobs according to the target feature. If list of jobs is not given
        and reset is True the attribute self.jobs is rewritten with the sorted list.

        Parameters
        ----------
        jobs_to_sort : (list), optional
            List of Job objects. If None self.jobs is used. The default is None.
        features : (str or list), optional
            Features to use to sort the list (see get_object_feature).
            Both a single variable or a list can be provided. The default is 'name'.
        reset : (bool), optional
            Reset the self.jobs attribute if jobs_to_sort is None. The default is True.
        reverse : (bool), optional
            Reverse the order in the list. The default is False.

        Returns
        -------
        sorted_jobs : (list)
            List of sorted Job objects.
        """
        if not isinstance(features,list):
            features = [features]
        jobs = jobs_to_sort if jobs_to_sort else self.jobs
        sorted_jobs = sort_objects(jobs, features, reverse)

        if not jobs_to_sort:
            if reset:
                self.jobs = sorted_jobs
                self.sort = features
                return
            else:
                return sorted_jobs
        else:
            return sorted_jobs


    def sync_dataset_from_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync Dataset project folder from HPC to local machine

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        hpc = HPCInterface()
        abs_path = op.abspath(self.path)
        localdir = abs_path 
        stdout,stderr = hpc.rsync_from_hpc(localdir=localdir,remotedir=self.path_in_hpc,exclude=exclude,dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return


    def sync_dataset_to_hpc(self,stdouts=False,exclude=None,dry_run=False):
        """
        Sync Dataset project folder from local machine to HPC

        Parameters
        ----------
        stdouts : (bool), optional
            Return output and error strings. The default is False.
        exclude : (list), optional
            List of files to exclude in rsync. The default is None.
        dry_run : (bool), optional
            Perform dry run in rsync with --dry-run. The default is False.

        Returns
        -------
        stdout : (str)
            Output.
        stderr : (str)
            Error.
        """
        hpc = HPCInterface()
        abs_path = op.abspath(self.path)
        localdir = abs_path 
        stdout,stderr = hpc.rsync_to_hpc(localdir=localdir,remotedir=self.path_in_hpc,exclude=exclude,dry_run=dry_run)
        if stdouts:
            return stdout,stderr
        else:
            return

           
    def sync_jobs(self):
        """Sync job data from HPC to local machine"""
        for j in self.jobs:
            j.sync_from_hpc()
        self.get_jobs_outputs()
        return
        
                
    def write_jobs_input(self):
        """Write jobs inputs to files"""
        for job in self.jobs:
            job.write_input()
            
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Sep  8 19:20:55 2023

@author: lorenzo
"""
import numpy as np

from pymatgen.io.vasp.inputs import Potcar


def get_charge_from_computed_entry(entry):
    potcar_symbols_bulk = [d['titel'].split(' ')[1] for d in entry.parameters['potcar_spec']]
    potcar = Potcar(potcar_symbols_bulk)
    charge = 0
    if 'parameters' in entry.data.keys():
        parameters = entry.data['parameters']
    else:
        raise ValueError('"parameters" need to be present in ComputedEntry data')
    if 'NELECT' in parameters.keys():
        nelect = parameters['NELECT']
        val = {}
        for p in potcar:
            val[p.element] = p.nelectrons
        neutral = sum([ val[el.symbol]*coeff 
                       for el,coeff in entry.structure.composition.items()])
        charge = neutral - nelect
    if not isinstance(charge,int):
        charge = np.around(charge,decimals=1)
        
    return charge


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 14:55:50 2020

@author: villa
"""


from matgendb.query_engine import QueryEngine
from pymatgen.io.vasp.inputs import VaspInput, Incar,Kpoints,Poscar,Potcar
from pynter.vasp.jobs import VaspJob
from pynter import SETTINGS


dbconfig = SETTINGS['dbconfig']

class VaspJobQuery(QueryEngine):
    """
    Subclass of QueryEngine in pymatgen-db (matgendb) package.
    To set up the database for VaspJob we use the pymatgen-db package which already provides a useful
    interface to the other pymatgen objects that are already used (like Structure or ComputedEntry).
    This class helps to interface a VaspJob object with the database.
    This class sets up the default "optional_data" to get from the db which allows to convert 
    pymatgen's ComputedStructureEntry objects retrieved from the db to VaspJob objects. 
    """
    def __init__(self,aliases_config=None,default_properties=None,query_post=None,
                 result_post=None,connection=None,replicaset=None,**kwargs):
        
        for k,v in dbconfig['vasp'].items():
            if k not in kwargs:
                kwargs[k] = v
        host = kwargs['host']
        port = kwargs['port']
        database = kwargs['database']
        user = kwargs['user']
        password = kwargs['password']
        collection = kwargs['collection']
        
        #can't pass kwargs directly because of **ignore in QueryEngine
        super().__init__(host=host,port=port,database=database,user=user,password=password,collection=collection,
                         aliases_config=aliases_config,default_properties=default_properties,query_post=query_post,
                         result_post=result_post,connection=connection,replicaset=replicaset) 
        
        self.optional_data = ['calculations','dir_name','final_energy','eigenvalue_band_properties',
                         'job_settings','job_script_filename','job_name','is_converged','structures']
        

    def convert_computed_entry_to_job(self,entry):
        """
        Convert ComputedStructureEntry into VaspJob object

        Parameters
        ----------
        entry : 
            ComputedStructureEntry.

        Returns
        -------
        vaspjob : 
            VaspJob object.
        """
        e = entry
        path = e.data['dir_name']
        
        inp = e.data['calculations'][0]['input']           
        incar = Incar(inp['incar'])
        kpoints = Kpoints.from_dict(inp['kpoints'])
        poscar = Poscar(e.structure)
        potcar = Potcar(inp['potcar'])
        
        inputs = VaspInput(incar, kpoints, poscar, potcar)
        job_settings = e.data['job_settings']
        job_script_filename = e.data['job_script_filename']
        name = e.data['job_name']
        
        outputs = {'ComputedStructureEntry':e}
        
        vaspjob =  VaspJob(path,inputs,job_settings,outputs,job_script_filename,name)
        vaspjob._is_converged = e.data['is_converged']
        vaspjob._band_structure = None
            
        return vaspjob            


    def get_entries_optional_data(self,criteria,optional_data=[]):
        """
        Get entries with get_entries method with default "optional_data".

        Parameters
        ----------
        criteria : 
            Criteria obeying the same syntax as query.
        optional_data : (list), optional
            Optional data to include in the entry. The default is [].

        Returns
        -------
        entries : (list)
            List of ComputedStructureEntry objects.
        """
        optional_data = optional_data + self.optional_data
        entries = self.get_entries(criteria,inc_structure=True,optional_data=optional_data)
        return entries


    def get_entries_in_system_optional_data(self,elements,optional_data=[],additional_criteria=None):
        """
        Uses get_entries_in_system method with default "optional_data".
        Gets all entries in a chemical system, e.g. Li-Fe-O will return all
        Li-O, Fe-O, Li-Fe, Li-Fe-O compounds.
        
        Parameters
        ----------
        elements: (list)
            List of strings with element symbols.
        optional_data : (list), optional
            Optional data to include in the entry. The default is [].
        criteria : 
            Added ability to provide additional criteria other than just
            the chemical system.

        Returns
        -------
        entries : (list)
            List of ComputedStructureEntry objects.
        """        
        optional_data = optional_data + self.optional_data
        entries = self.get_entries_in_system(elements,inc_structure=True, optional_data=optional_data,
                                             additional_criteria=additional_criteria)
        return entries
    
    
    def get_jobs(self,criteria,optional_data=[]):
        """
        Get VaspJob objects from db. First retrives ComputedStructureEntry objects than 
        converts them into VaspJob objects.

        Parameters
        ----------
        criteria : 
            Criteria obeying the same syntax as query.
        optional_data : (list), optional
            Optional data to include in the entry. The default is [].

        Returns
        -------
        jobs : (list)
            List of VaspJob objects.
        """
        optional_data = optional_data + self.optional_data
        jobs = []
        entries = self.get_entries_optional_data(criteria,optional_data=optional_data)
        for e in entries:
            j = self.convert_computed_entry_to_job(e)
            jobs.append(j)
        
        return jobs


            


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 26 14:04:04 2023

@author: villa
"""

import os
import json

from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.electronic_structure.plotter import DosPlotter, BSDOSPlotter, BSPlotter 
from pymatgen.electronic_structure.dos import CompleteDos
from pymatgen.analysis.transition_state import NEBAnalysis


def setup_plotter(subparsers):
    
    parser_plot = subparsers.add_parser('plot',help='Plotting tools for DOS, BS and NEB')
    subparsers_plot = parser_plot.add_subparsers()
    
    parser_bs = subparsers_plot.add_parser('BS',help='Plot band structure with pymatgen')
    setup_plot_bs(parser_bs)
    
    parser_dos = subparsers_plot.add_parser('DOS',help='Plot density of states with pymatgen')
    setup_plot_dos(parser_dos)
    
    parser_dos_bs = subparsers_plot.add_parser('DOS-BS',help='Plot density of states and band structure with pymatgen')
    setup_plot_dos_bs(parser_dos_bs)
    
    parser_neb = subparsers_plot.add_parser('NEB',help='Plot NEB energy landscape with pymatgen')
    setup_plot_neb(parser_neb)
    
    return
    



def setup_plot_bs(parser_sub):    
    parser_sub.add_argument('-k','--kpoints-filename',help='KPOINTS filename',required=False,type=str,default=None,metavar='',dest='kpoints_filename')    
    parser_sub.add_argument('-s','--save',help='Save fig as pdf',required=False,default=False,action='store_true',dest='savefig')   
    parser_sub.add_argument('-hy','--hybrid-mode',help='Force hybrid mode for BS',required=False,default=False,action='store_true',dest='force_hybrid_mode') 
    parser_sub.add_argument('-l','--line-mode',help='Line mode for BS',required=False,default=True,action='store_true',dest='line_mode')   
    parser_sub.add_argument('-y','--ylim',help='Range for y-axis',required=False,default=None,nargs='+',type=float,metavar='',dest='ylim')
    parser_sub.add_argument('-sm','--smooth',help='Smooth band structure',required=False,default=False,action='store_true',dest='smooth')  
    parser_sub.add_argument('-t','--title',help='Title of the plot',required=False,type=str,default=None,metavar='',dest='title') 
    parser_sub.set_defaults(func=plot_bs) 
    return


def plot_bs(args):
    v = Vasprun("vasprun.xml")
    bs = v.get_band_structure(kpoints_filename=args.kpoints_filename,line_mode=args.line_mode,
                              force_hybrid_mode=args.force_hybrid_mode)
    plt = BSPlotter(bs).get_plot(ylim=args.ylim,smooth=args.smooth)
    plt.gca().get_legend().remove()
    
    if args.title:
        plt.title(args.title,fontdict={'fontsize':25})
    if args.savefig:
        plt.savefig('BS.pdf',bbox_inches='tight')
    else:
        plt.show()
    return    
    




def setup_plot_dos(parser_sub): 
    parser_sub.add_argument('-sg','--sigma',help='Standard deviation for Gaussian smearing of DOS',required=False,type=float,default=None,metavar='',dest='sigma')    
    parser_sub.add_argument('-s','--save',help='Save fig as pdf',required=False,default=False,action='store_true',dest='savefig')   
    parser_sub.add_argument('-st','--stack',help='Stack DOS',required=False,default=False,action='store_true',dest='stack') 
    parser_sub.add_argument('-y','--ylim',help='Range for y-axis',required=False,default=None,nargs='+',type=float,metavar='',dest='ylim')
    parser_sub.add_argument('-x','--xlim',help='Range for x-axis',required=False,default=None,nargs='+',type=float,metavar='',dest='xlim')
    parser_sub.add_argument('-p','--projection',help='Projection of DOS, "elements", "orbitals" or "None"',required=False,type=str,default='elements',metavar='',dest='dos_projection')    
    parser_sub.add_argument('-t','--title',help='Title of the plot',required=False,type=str,default=None,metavar='',dest='title') 
    parser_sub.set_defaults(func=plot_dos)
    return


def plot_dos(args):
    xlim = args.xlim if args.xlim else (-10,10)
    complete_dos = Vasprun('vasprun.xml').complete_dos    
    plotter = DosPlotter(stack=args.stack,sigma=args.sigma)
    plotter.add_dos('Total',complete_dos)
    
    if args.dos_projection == 'orbitals':
        dos_dict = complete_dos.get_spd_dos()
    elif args.dos_projection == 'elements':
        dos_dict = complete_dos.get_element_dos()
    for k in dos_dict:
        if not args.dos_projection=='None':   
            plotter.add_dos(k,dos_dict[k])
    plt = plotter.get_plot(xlim=xlim,ylim=args.ylim)
    plt.title(args.title)
    if args.savefig:
        plt.savefig('DOS.pdf',bbox_inches='tight')
    else:
        plt.show()
        
 
    
 
        
def setup_plot_dos_bs(parser_sub):    
    parser_sub.add_argument('-p','--path-to-dos',help='Path to Dos object saved as json or path to where vasprun for dos is stored',required=False,type=str,default=None,metavar='',dest='path_to_dos')    
    parser_sub.add_argument('-s','--save',help='Save fig as pdf',required=False,default=False,action='store_true',dest='savefig')   
    parser_sub.add_argument('-hy','--hybrid-mode',help='Force hybrid mode for BS',required=False,default=False,action='store_true',dest='hybrid')   
    parser_sub.add_argument('-y','--ylim',help='Range for y-axis',required=False,default=None,nargs='+',type=float,metavar='',dest='ylim')
    parser_sub.add_argument('-t','--title',help='Title of the plot',required=False,type=str,default=None,metavar='',dest='title') 
    parser_sub.set_defaults(func=plot_dos_bs)
    return


def plot_dos_bs(args):
    if not args.ylim:
        args.ylim = (-10,10)    
    get_dos = False
    if args.path_to_dos:
        if '.json' in args.path_to_dos:
            with open(args.path_to_dos) as f:
                dos = CompleteDos.from_dict(json.load(f))
        elif args.path_to_dos:
            try:
                dos = Vasprun(os.path.join(args.path_to_dos,"vasprun.xml")).complete_dos
            except:
                raise ValueError(f'Reading of dos from vasprun.xml in {args.path_to_dos} failed')
    else:
        get_dos = True    
    v = Vasprun('vasprun.xml')    
    if get_dos:
        dos = v.complete_dos
    efermi = dos.efermi
    gap = dos.get_gap()
    ylim = args.ylim
    vb_range = -1*ylim[0] if ylim else 4  # adjust for nonsense args in BSDOSPlotter
    cb_range = ylim[1] - gap if ylim else 4  # adjust for nonsense args in BSDOSPlotter
    bs = v.get_band_structure(line_mode=True,force_hybrid_mode=args.hybrid,efermi=efermi)
    
    plt = BSDOSPlotter(bs_projection=None, dos_projection='elements', bs_legend=None, 
                        vb_energy_range=vb_range, cb_energy_range=cb_range).get_plot(bs,dos)
    ax = plt.gcf().get_axes()[0]
    ax.set_ylabel('$E - E_F$ (eV)')    
    if args.savefig:
        plt.savefig('DOS-BS.pdf')
    else:
        plt.show()
    
    
    
    
    
def setup_plot_neb(parser_sub):   
    parser_sub.add_argument('-p','--path',help='Path to NEB calculation',required=False,type=str,default=None,metavar='',dest='path')
    parser_sub.add_argument('-pr','--print',help='Print energies and forces',required=False,default=False,action='store_true',dest='print_energies')
    parser_sub.add_argument('-s','--save',help='Save fig as pdf',required=False,default=False,action='store_true',dest='savefig')
    parser_sub.add_argument('-t','--title',help='Title of the plot',required=False,type=str,default=None,metavar='',dest='title')
    parser_sub.set_defaults(func=plot_neb)
    return
    
    
def plot_neb(args):
    path = args.path if args.path else os.getcwd()
    path = os.path.abspath(path)
    neb_analysis = NEBAnalysis.from_dir(path)
    if args.print_energies:
        print('Energies:\n',neb_analysis.energies)
        print('Forces:\n',neb_analysis.forces)
    plt = neb_analysis.get_plot()
    plt.title(args.title)
    if args.savefig:
        plt.savefig('NEB.pdf')
    else:
        plt.show()
        
    
    
    
    
    
    
    
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 16 14:35:23 2021

@author: villa
"""

import numpy as np
import pandas as pd
import matplotlib

class Stresses:
    
    def __init__(self,stress_bulk):
        """
        Class to handle elastic analysis from defect calculations. Necessary data must be present for 
        defect entries in dictionary form in "data" attribute.
        Valid keys are:
            "stress" : To compute elastic dipole tensor and relaxation volume starting from stress tensor 
                    of defect calculation.
            "elastic_corrections" : dict with corrections for charged defects ({"corr_name":corr_value}).
                    This values will be added to the elastic dipole tensor.

        Parameters
        ----------
        stress_bulk : (np.array)
            Stress tensor of bulk calculation.
        """
        self.stress_bulk = stress_bulk
        
        
    def get_elastic_dipole_tensor(self,entry,add_corrections=True,add_to_data=False):
        """
        Calculate elastic dipole tensor as P_ij = -Omega * sigma_ij, where Omega is the 
        bulk volume and sigma_ij is the residual stress tensor.

        Parameters
        ----------
        entry : (DefectEntry)
            DefectEntry object. Stress needs to be present in the data
            dictionary as a numpy array with key "stress".
        add_corrections : (bool)
            Add correction terms from "elastic_corrections" dict (if key is present in dict). The default is True.
 
        Returns
        -------
        dipole_tensor : (np.array)
            Elastic dipole tensor in eV.
        """
        bulk_volume = entry.bulk_structure.lattice.volume
        res_stress = self.get_residual_stress_tensor(entry)
        dipole_tensor = -1*bulk_volume*res_stress #sign is inverted with respect to VASP output
        if 'elastic_corrections' in entry.data.keys() and add_corrections:
            dipole_tensor += sum([v for k,v in entry.data['elastic_corrections'].items()]) 
        if add_to_data:
            entry.data['elastic_dipole_tensor'] = dipole_tensor
        return dipole_tensor


    def get_relaxation_volume(self,entry,bulk_modulus,add_corrections=True,add_to_data=True):
        """
        Calculate relaxation volume from stresses. Stresses data needs to be in numpy.array format and present 
        in the "data" dictionary with relative "stress" key.

        Parameters
        ----------
        entry : (DefectEntry)
            DefectEntry object. Stress needs to be present in the data dictionary 
            as a numpy array with key "stress".  
        bulk_modulus : (float)
            Bulk modulus in GPa.
        add_corrections : (bool)
            Add correction terms from "elastic_corrections" dict (if key is present in dict). The default is True.
        add_to_data : (bool)
            Add relaxation volume value to entry.data dict with key "relaxation_volume". The default is True.
            
        Returns
        -------
        rel_volume : (float)
            Relaxation volume in A°^3.
        """
        bulk_modulus = bulk_modulus*10 # from GPa to kbar
        dipole_tensor = self.get_elastic_dipole_tensor(entry)
        pressure = np.trace(dipole_tensor)/3
        rel_volume = pressure/bulk_modulus
        if add_to_data:
            entry.data['relaxation_volume'] = rel_volume
        return rel_volume

        
    def get_residual_stress_tensor(self,entry): 
        """
        Calculate residual stress tensor by subtracting bulk stress tensor.

        Parameters
        ----------
        entry : (DefectEntry)
            DefectEntry object. Stress needs to be present in the data dictionary 
            as a numpy array with key "stress".

        Returns
        -------
        stress : (np.array)
            Residual stress matrix.
        """ 
        if 'stress' in entry.data.keys():
            stress_def = entry.data['stress']
            stress = np.array(stress_def) - np.array(self.stress_bulk) # residual stress -> defect - bulk
            return stress
        else:
            raise UserWarning('Stress needs to be present in the data dictionary as a numpy array with key "stress".')
            return
    
        
    def plot_relaxation_volumes(self,entries,bulk_modulus,check_in_data=True,add_corrections=True,
                                add_to_data=True,get_dataframe=False,**kwargs):
        """
        Plot relaxation volumes for list of defect entries with pandas Series.

        Parameters
        ----------
        entries : (list)
            List of DefectEntry objects.
        bulk_modulus : (float)
            Bulk modulus in GPa.
        check_in_data : (bool), optional
            Check weather relaxation volume is already present in entry.data['relaxation_volume']. The default is True.
        add_corrections : (bool)
            Add correction terms from "elastic_corrections" dict (if key is present in dict). The default is True.
        add_to_data : (bool)
            Add relaxation volume value to entry.data dict with key "relaxation_volume". The default is True.
        get_dataframe : (bool), optional
            Return tuple with pandas data: (plt,dataframe). If False only the matplotlib object is returned.
            The default is False.
        **kwargs : (dict)
            Kwargs for df.plot method.

        Returns
        -------
        (matplotlib object)
            If get_dataframe is True return Tuple with matplotlib object and pandas data: (plt,dataframe). 
            If False only the matplotlib object is returned.
        """
        rel_volumes = {}
        for e in entries:
            if check_in_data and 'relaxation_volume' in e.data.keys():
                rel_volumes[e.symbol] = e.data['relaxation_volume']                  
            else:
                rel_volumes[e.symbol] = self.get_relaxation_volume(e,bulk_modulus,add_corrections,add_to_data)
                
        df = pd.Series(rel_volumes)
        if not kwargs:
            kwargs = {'grid':True,'figsize':(12,8)} #default values
        fontsize = 15*kwargs['figsize'][0]/12 
        matplotlib.rcParams.update({'font.size':fontsize})
        plt = df.plot.bar(rot=0,ylabel='Relaxation Volume ($\AA^{3}$)',**kwargs).get_figure()
        
        if get_dataframe:
            return plt,df
        else:
            return plt
        
    
    
    
    
    